# coding: utf-8
"""Classes to analyse LRUJ results."""
from __future__ import annotations

import dataclasses
import numpy as np
import pandas as pd
from pathlib import Path
#from typing import Any
#from monty.string import is_string, list_strings, marquee
from abipy.core.mixins import AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.iotools import ETSF_Reader
from abipy.tools.iotools import yaml_safe_load
from abipy.core.structure import Structure
from abipy.tools.typing import Figure, PathLike
from abipy.tools.plotting import (set_axlims, add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt,
    get_ax3d_fig_plt, rotate_ticklabels, set_visible, plot_unit_cell, set_ax_xylabels, get_figs_plotly)



#class LrujFile(AbinitNcFile, Has_Header, Has_Structure): #, Has_ElectronBands, NotebookWriter):
#    """
#    File containing the results of a ground-state calculation.
#
#    Usage example:
#
#    .. code-block:: python
#
#        with GsrFile("foo_GSR.nc") as gsr:
#            print("energy: ", gsr.energy)
#            gsr.ebands.plot()
#
#    .. rubric:: Inheritance Diagram
#    .. inheritance-diagram:: LrujFile
#    """
#
#    @classmethod
#    def from_file(cls, filepath: str) -> GsrFile:
#        """Initialize the object from a netcdf_ file."""
#        return cls(filepath)
#
#    def __init__(self, filepath: str):
#        super().__init__(filepath)
#        self.r = r = EtsfReader(filepath)


#===============================================================================================================
#===============================================================================================================
@dataclasses.dataclass(kw_only=True)
class LrujResults:
    """
    This object stores the results produced by lruj.
    """
    natom: int
    npert: int
    ndata: int
    pawujat: int
    macro_uj: int
    diem_token: str
    diem: float
    alphas: np.ndarray
    occ_unscr: np.ndarray
    occ_scr: np.ndarray
    chi0_coefficients: dict
    chi_coefficients: dict
    maxdeg: int
    dmatpuopt: int
    pert_name: str
    parname: str
    metric: str
    fit_df: pd.DataFrame

    @classmethod
    def from_file(cls, filepath: PathLike):
        """
        Extract results from the main ouput file produced by lruj.
        """
        with open(filepath, "rt") as fh:
            lines = [line.lstrip() for line in fh]

        # Extract the Yaml document with the chi/chi0 coefficients
        in_doc = False
        yaml_lines = []
        for i, line in enumerate(lines):

            if line.startswith("--- !LRUJ_Abipy_Plots"):
                in_doc = True
                continue

            if in_doc and line.startswith("..."):
                data = yaml_safe_load("".join(yaml_lines))
                break

            if in_doc:
                yaml_lines.append(line)

        natom = data['natom']
        ndata = data['ndata']
        pawujat = data['pawujat']
        macro_uj = data['macro_uj']
        diem_token = data['diem_token']
        diem = data['diem']
        npert = ndata - 1
        if macro_uj==4:
          pert_name = r"$\beta$"
          metric = r"M $(n^{\uparrow} - n^{\downarrow})$"
          parname = 'J'
        else:
          pert_name = r"$\alpha$"
          metric = r"N $(n^{\uparrow} + n^{\downarrow})$"
          parname = 'U'

        chi0_coefficients = {}
        chi_coefficients = {}
        for k, v in data.items():
            magic = "chi0_coefficients_degree"
            if k.startswith(magic):
                degree = int(k.replace(magic, ""))
                chi0_coefficients[degree] = v
            magic = "chi_coefficients_degree"
            if k.startswith(magic):
                degree = int(k.replace(magic, ""))
                chi_coefficients[degree] = v

        #print(f"{chi_coefficients=}")

        def find(header, dtype=None):
            for i, line in enumerate(lines):
                if line.startswith(header):
                    after = line.replace(header, "", 1).strip()
                    if dtype: after = dtype(after)
                    return i, after
            raise ValueError(f"Cannot find {header=} in {filepath=}")

        _, maxdeg = find("Maximum degree of polynomials analyzed:", dtype=int)
        _, dmatpuopt = find("Value of dmatpuopt:", dtype=int)

        # Parse the section with perturbations and occupations.
        """
        Perturbations           Occupations
        -------------- -----------------------------
          alpha [eV]     Unscreened      Screened
        -------------- -----------------------------
         0.0000000000   8.6380182458   8.6380182458
        -0.1500000676   8.6964981922   8.6520722003

       -OR-

        Perturbations         Magnetizations
        --------------- -----------------------------
           beta [eV]     Unscreened      Screened
        --------------- -----------------------------

        """
        i, _ = find("Perturbations",dtype=None)
        i += 4
        vals = []
        for ipert in range(ndata):
            vals.append([float(t) for t in lines[i+ipert].split()])
        vals = np.reshape(vals, (ndata, 3))
        alphas, occ_unscr, occ_scr = vals[:,0], vals[:,1], vals[:,2]
        """
                                                                               RMS Errors
                                                                 ---------------------------------------
         Regression   Chi0 [eV^-1]   Chi [eV^-1]      J [eV]    | Chi0 [eV^-1]  Chi [eV^-1]     J [eV]
        --------------------------------------------------------|---------------------------------------
         Linear:        -0.8594082    -0.0949434     9.3689952  |    0.0023925    0.0000878    0.1139297
         Quadratic:     -0.8574665    -0.0955791     9.2963099  |    0.0023777    0.0000129    0.0681722
         Cubic:         -0.8007858    -0.0952726     9.2474220  |    0.0001546    0.0000015    0.0200543
        *************************************************************************************************
        """
        i, _ = find("Regression   Chi0 [eV^-1]")
        i += 2
        keys = ["Chi0", "Chi", "HP", "rms_Chi0", "rms_Chi", "rms_HP"]
        dict_list = []
        for irow in range(maxdeg):
            l = lines[i+irow].replace("|", " ")
            tokens = l.split()
            d = dict(zip(keys, [float(t) for t in tokens[-6:]]))
            d["degree"] = irow + 1
            dict_list.append(d)

        fit_df = pd.DataFrame(dict_list)

        # Build instance from locals dict.
        _data = locals()
        return cls(**{k: _data[k] for k in [field.name for field in dataclasses.fields(cls)]})

#===============================================================================================================
#===============================================================================================================
    @add_fig_kwargs
    def plot(self, ax=None, degrees="all", inset=True, insetdegree=1, insetlocale="lower left",
             ptcolor0='k', ptcolor='k', gradcolor1='#3575D5',gradcolor2='#FDAE7B',
             ptitle="default", fontsize=12, **kwargs) -> Figure:
        """
        Plot

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            degrees: List of degrees to plot. If None, no polynomials are plotted.
            ptcolor0: Color of unscreened response data points (default: black)
            ptcolor: Color of screened response data points (default: black)
            gradcolor1: Hex code of linear regression color (default: Blue #3575D5)
            gradcolor2: Hex code of color of regression of maximum degree in list <degrees> (default: Salmon #FDAE7B)
            ptitle: Plot title (default: "Linear Response for atom <pawujat>)
            inset: Plots inset with LR parameter for polynomial fit of degree <insetdegree> (default: True)
            insetdegree: Degree of polynomial fit information printed in the inset (default: 1)
            insetlocale: Position of inset in the plot. Standard matplotlob locations. (default: "lower left")
        """
        import seaborn as sns
        sns.set()
        ax, fig, plt = get_ax_fig_plt(ax=ax)

        # Plot data
        yshift = self.occ_unscr[np.where(self.alphas == 0.0000)] * (self.diem - 1.0)
        data0 = 1.0/self.diem * (self.occ_unscr + yshift)
        ax.scatter(self.alphas, data0, s=70, color=ptcolor0, facecolors='none', linewidths=2, label="Unscreened")
        ax.scatter(self.alphas, self.occ_scr, s=70, color=ptcolor, label="Screened")
        ax.axvline(x=0.0, color="white", linestyle="--", lw=0.5)

        # Generate mesh for polynomial functions
        xstart, xstop = 1.1 * self.alphas.min(), 1.1 * self.alphas.max()
        xs = np.arange(xstart, xstop, step=0.01)

        # Prepare colors and coefficients for polynomials the use wants to plot
        if degrees == "all":
          degrees = self.chi0_coefficients.keys()

        def hex_to_RGB(hex_str):
          return [int(hex_str[i:i+2], 16) for i in range(1,6,2)]

        def get_color_gradient(c1, c2, n):
          assert n > 1
          c1_rgb = np.array(hex_to_RGB(c1))/255
          c2_rgb = np.array(hex_to_RGB(c2))/255
          mix_pcts = [x/(n-1) for x in range(n)]
          rgb_colors = [((1-mix)*c1_rgb + (mix*c2_rgb)) for mix in mix_pcts]
          return ["#" + "".join([format(int(round(val*255)), "02x") for val in item]) for item in rgb_colors]

        linecolors=get_color_gradient(gradcolor1,gradcolor2,len(degrees))

        # Plot polynomial functions
        def poly0(coeffs):
          return lambda x: 1.0/self.diem * (sum((coeff*x**i for i,coeff in enumerate(coeffs))) + yshift)

        def poly(coeffs):
          return lambda x: sum((coeff*x**i for i,coeff in enumerate(coeffs)))

        for degree in degrees:
          polynomial0=poly0(self.chi0_coefficients[degree])
          polynomial=poly(self.chi_coefficients[degree])
          if degree == 1:
            Labelstring='Linear'
          elif degree == 2:
            Labelstring='Quadratic'
          elif degree == 3:
            Labelstring='Cubic'
          else:
            Labelstring=' '.join(['Degree',str(degree)])

          if insetdegree==degree:
            deginfo = ' '.join(['Parameters for',Labelstring,'fit'])
            insetcolor = linecolors[degree-1]

          ax.plot(xs,polynomial0(xs),color=linecolors[degree-1],linewidth=2.0,linestyle='dashed')
          ax.plot(xs,polynomial(xs),color=linecolors[degree-1],linewidth=2.0,label=Labelstring)

        ax.legend(loc="best", fontsize=fontsize, shadow=True)
        if ptitle=="default":
          ptitle=' '.join(["Linear Response",self.parname,"on atom",str(self.pawujat)])
        plt.title(ptitle)
        ax.grid(True)
        ax.set_xlabel(' '.join([self.pert_name,"(eV)"]))
        ax.set_ylabel(self.metric)

        # Generate inset with numerical information on LR
        if inset:
          from matplotlib.offsetbox import AnchoredText
          select =  self.fit_df["degree"] == insetdegree
          def dfvalue(keywo):
            tablevalue = self.fit_df[select][keywo]
            return "%.4f" % tablevalue.values[0]

          X0str = ' '.join([r"$\chi_0$","=",dfvalue("Chi0"),r"$\pm$",dfvalue("rms_Chi0"),r"[eV]$^{-1}$"])
          Xstr = ' '.join([r"$\chi$","=",dfvalue("Chi"),r"$\pm$",dfvalue("rms_Chi"),r"[eV]$^{-1}$"])
          HPstr = ' '.join([self.parname,"=",dfvalue("HP"),r"$\pm$",dfvalue("rms_HP"),r"[eV]"])
          insettxt = '\n'.join([deginfo,X0str,Xstr,HPstr])
          parambox = AnchoredText(insettxt,loc=insetlocale)
          parambox.patch.set_linewidth(2)
          parambox.patch.set_edgecolor(insetcolor)
          parambox.patch.set_facecolor('white')
          ax.add_artist(parambox)

        return fig



#===============================================================================================================
#===============================================================================================================
class LrujAnalyzer:
    """
    Analyzes multiple sets of LRUJ files.
    """
    def __init__(self, manager=None, verbose=0):
        self.ncfiles_of_key = {}
        self.results_of_key = {}
        self.manager = manager
        self.verbose = verbose

    #def __str__(self):
    #    return self.to_string()

    #def to_string(self, verbose: int = 0) -> str:
    #    lines = []
    #    app = lines.append
    #    return "\n".join(lines)

    #@classmethod
    #def from_dir(cls, key: str, directory: PathLike) -> None:
    #    new = cls()
    #    new.scan_dir(key, directory)
    #    return new

    #def walk_dir(self, key: str, top: PathLike) -> None:
    #    nc_paths = None
    #    self.add_ncpaths(key, nc_paths)

    def add_ncpaths(self, key: str, nc_paths: list[str]) -> None:
        self.ncfiles_of_key[key] = nc_paths
        self.results_of_key[key] = None
        self.run()

    def run(self, workdir=None) -> None:
        """
        Invoke lruj
        """
        from abipy.flowtk.wrappers import Lruj
        from abipy.core.globals import get_workdir
        workdir = Path(get_workdir(workdir))
        for key, nc_paths in self.ncfiles_of_key.items():
            if self.results_of_key[key] is not None: continue
            wdir = workdir / f"run_{key}"
            wdir.mkdir()
            Lruj(manager=self.manager, verbose=self.verbose).run(nc_paths, workdir=wdir)
            self.results_of_key[key] = LrujResults.from_file(wdir / "lruj.stdout")

    @add_fig_kwargs
    def plot(self, **kwargs) -> Figure:
        """
        Plot

        Args:
        """
        keys = self.results_of_key.keys()

        # Build grid of plots.
        num_plots, ncols, nrows = len(keys), 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=True, squeeze=False)
        ax_list = ax_list.ravel()
        # don't show the last ax if num_plots is odd.
        if num_plots % ncols != 0: ax_list[-1].axis("off")

        for ix, (key, ax) in enumerate(zip(keys, ax_list)):
            res = self.results_of_key[key]
            res.plot(ax=ax, show=False, title=key)

        return fig


#class LrujInputGenerator
#
#    def __init__(self, gs_input, elements, site_inds=None, dmatpuopt=3):
#        self.dmatpuopt = dmatpuopt
#        self.gs_input = gs_input.copy()
#        self.gs_input.set_vars(
#            chkprim=0,            # Will complain otherwise with AFII magnetic state
#            chksymbreak=0,        # Don't check for symmetry breaking
#            # DFT+U
#            usepawu=1,            # Alert Abinit to use of DFT+U
#            lpawu=[2, 2, 1],      # Subspaces of interest: Ni 3d, O 2p
#            upawu="*0.0",         # Raw (non-corrected) XC functional required to establish U(J)
#            jpawu="*0.0",
#            dmatpuopt=self.dmatpuopt, # PAW projector for density matrix
#        )
#
#    def gen_inputs(self, macro_uj, alphas, supercells):
#        #supercells = np.respahe(supercells, (-1,
#        for supercell in supercells:
#            for alpha in alphas:
#                if alpha == 0.0: continue
#                gs_input.new_with_vars(
#                    macro_uj=macro_uj,
#                    pawujat=1,              # Label of perturbed atom
#                    pawujv1=f"{alpha} eV",  # List of perturbation strengths
#                    prtwf=0,
#                    prtebands 0             # Don't print ebands
#                )



#!/usr/bin/env python
"""Benchmark for k-point parallelism."""
import sys
import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata

from itertools import product
from abipy.benchmarks import bench_main, BenchmarkFlow


def make_input(paw=False):
    """Build a template for GS calculations with k-point parallelism"""
    pseudos = abidata.pseudos("14si.pspnc") if not paw else abidata.pseudos("Si.GGA_PBE-JTH-paw.xml")
    structure = abidata.structure_from_ucell("Si")

    inp = abilab.AbinitInput(structure, pseudos)
    inp.set_kmesh(ngkpt=[12,12,12], shiftk=[0,0,0])

    # Global variables
    ecut = 40
    inp.set_vars(
        ecut=ecut,
        pawecutdg=ecut*4 if paw else None,
        nsppol=2,
        nband=40,
        paral_kgb=0,
        #istwfk="*1",
        timopt=-1,
        chksymbreak=0,
        prtwf=0,
        prtden=0,
        tolvrs=1e-10,
    )

    return inp


def build_flow(options):
    inp = make_input(paw=options.paw)

    mpi_list = options.mpi_list
    if mpi_list is None:
        nkpt = len(inp.abiget_ibz().points)
        nks = nkpt * inp["nsppol"]
        mpi_list = [p for p in range(1, nks+1) if nks % p == 0]
        #print("Using mpi_list:", mpi_list)
    else:
        print("Using mpi_list from cmd line:", mpi_list)

    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)
    work = flowtk.Work()

    for mpi_procs, omp_threads in product(mpi_list, options.omp_list):
        if not options.accept_mpi_omp(mpi_procs, omp_threads): continue
        manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
        work.register_scf_task(inp, manager=manager)

    flow.register_work(work)
    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return

    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
"""
This script retrieve information on Slurm jobs.
"""
import sys
import os
import argparse
import abipy.tools.cli_parsers as cli
import abipy.flowtk.qutils as qu

from abipy.core.release import __version__


def get_epilog() -> str:
   return """\
Usage example:\n

    abislurm.py running                => Get info on all the running jobs
    abislurm.py completed 111 112      => Get info on completed jobs
"""


def get_parser(with_epilog=False):
    # Build the main parser.
    parser = argparse.ArgumentParser(epilog=get_epilog() if with_epilog else "",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-v', '--verbose', default=0, action='count', # -vv --> verbose=2
                        help='verbose, can be supplied multiple times to increase verbosity')

    #parser.add_argument('-V', '--version', action='version', version="%(prog)s version " + __version__)
    #parser.add_argument('--loglevel', default="ERROR", type=str,
    #                    help="set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")

    # Parent parser for common options.
    copts_parser = argparse.ArgumentParser(add_help=False)
    copts_parser.add_argument('-v', '--verbose', default=0, action='count', # -vv --> verbose=2
        help='verbose, can be supplied multiple times to increase verbosity')
    copts_parser.add_argument('--loglevel', default="ERROR", type=str,
        help="Set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")

    job_ids_parser = argparse.ArgumentParser(add_help=False)
    job_ids_parser.add_argument('job_ids', nargs="+", help="List of job ids.")

    # Create the parsers for the sub-commands
    subparsers = parser.add_subparsers(dest='command', help='sub-command help',
        description="Valid subcommands, use command --help for help")

    # Subparser for running command.
    p_running = subparsers.add_parser('running', parents=[copts_parser],
        help="Check info on all the running jobs.")

    # Subparser for completed command.
    p_completed = subparsers.add_parser('completed', parents=[copts_parser, job_ids_parser],
        help="Returning info on completed jobs.")

    return parser


def main():

    def show_examples_and_exit(err_msg=None, error_code=1):
        """Display the usage of the script."""
        sys.stderr.write(get_epilog())
        if err_msg: sys.stderr.write("Fatal Error\n" + err_msg + "\n")
        sys.exit(error_code)

    parser = get_parser(with_epilog=True)

    # Parse command line.
    try:
        options = parser.parse_args()
    except Exception as exc:
        show_examples_and_exit(error_code=1)

    if not options.command:
        show_examples_and_exit(error_code=1)

    cli.set_loglevel(options.loglevel)

    if options.verbose > 2:
        print(options)

    if options.command == "running":
        jobs_dict = qu.slurm_get_jobs()
        for job_id, dct in jobs_dict.items():
            print(f"{job_id=}", dct)

    #elif options.command == "running_from_abilogs":

    elif options.command == "completed":
        for job_id in options.job_ids:
            print(qu.get_completed_job_info(job_id))

    #elif options.command == "completed_from_abilogs":
    #    job_ids = []

    else:
        raise ValueError(f"Unsupported {options.command=}")

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Function1D object
=================

This example shows how to use the Function1D object
to analyze and plot results.
"""
import numpy as np
import matplotlib.pyplot as plt

from abipy.abilab import Function1D

# Build mesh [0, 2pi] with 100 points.
mesh = np.linspace(0, 2*np.pi, num=100)

# Compute sine function.
sine = Function1D.from_func(np.sin, mesh)

# Call matplotlib to plot data.
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)

# Plot sine.
sine.plot_ax(ax, label="sin(x)")

# Spline sine on the coarse mesh sx, and plot the data.
sx = np.linspace(0, 2*np.pi, num=25)
splsine = sine.spline(sx)

plt.plot(sx, splsine, "ro", label="splined sine")

# Compute the 1-st and 2-nd order derivatives
# with finite differences (5-point stencil).
for order in [1,2]:
    der = sine.finite_diff(order=order)
    der.plot_ax(ax, label="D sin(x) %d" % order)

# Integrate the sine function and plot the results.
(1 - sine.integral()).plot_ax(ax, marker=".", label=r"$1 - \int\,\sin(x)\,dx$")

plt.legend(loc="best")
plt.show()



# coding: utf-8
"""Fold2Bloch netcdf file."""
from __future__ import annotations

import os
import numpy as np
import pymatgen.core.units as units

from monty.string import marquee
from monty.functools import lazy_property
from monty.collections import dict2namedtuple
from monty.termcolor import cprint
from pymatgen.core.lattice import Lattice
from abipy.core.structure import Structure
from abipy.core.mixins import AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.core.kpoints import KpointList, is_diagonal, find_points_along_path
from abipy.core.globals import get_workdir
from abipy.tools.plotting import set_axlims, add_fig_kwargs, get_ax_fig_plt
from abipy.tools.typing import Figure
from abipy.electrons.ebands import ElectronBands, ElectronsReader
from abipy.tools.numtools import gaussian


class Fold2BlochNcfile(AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter):
    """
    Netcdf file with output data produced by Fold2Bloch.

    Usage example:

    .. code-block:: python

        with Fold2BlochNcfile("foo_FOLD2BLOCH.nc") as fb:
            fb.plot_unfolded()

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: Fold2BlochNcfile
    """
    @classmethod
    def from_wfkpath(cls, wfkpath, folds, workdir=None, manager=None, mpi_procs=1, verbose=0) -> Fold2BlochNcfile:
        """
        Run fold2bloch in workdir.

        Args:
            wfkpath:
            folds:
            workdir: Working directory of the fake task used to compute the ibz. Use None for temporary dir.
            manager: :class:`TaskManager` of the task. If None, the manager is initialized from the config file.
            mpi_procs: Number of MPI processors to use.
            verbose: Verbosity level.
        """
        # Build a simple manager to run the job in a shell subprocess
        from abipy import flowtk
        manager = flowtk.TaskManager.as_manager(manager).to_shell_manager(mpi_procs=mpi_procs)
        fold2bloch = flowtk.Fold2Bloch(manager=manager, verbose=verbose)

        # Create temporary directory and link to the WFK file
        workdir = get_workdir(workdir)
        wfkpath = os.path.abspath(wfkpath)
        link = os.path.join(workdir, os.path.basename(wfkpath))
        os.symlink(wfkpath, link)

        # Run fold2bloch
        ncpath = fold2bloch.unfold(link, folds, workdir=workdir)

        return cls(ncpath)

    def __init__(self, filepath: str):
        super().__init__(filepath)
        self.reader = ElectronsReader(filepath)

        # Initialize the electron bands from file.
        # Spectral weights are dimensioned with `nss`
        # Fortran arrays.
        # nctkarr_t("reduced_coordinates_of_unfolded_kpoints", "dp", "number_of_reduced_dimensions, nk_unfolded")
        # nctkarr_t("unfolded_eigenvalues", "dp", "max_number_of_states, nk_unfolded, number_of_spins")
        # nctkarr_t("spectral_weights", "dp", "max_number_of_states, nk_unfolded, nsppol_times_nspinor")
        self._ebands = self.reader.read_ebands()
        self.nss = max(self.nsppol, self.nspinor)
        self.fold_matrix = self.reader.read_value("fold_matrix")

        # Compute direct lattice of the primitive cell from fold_matrix.
        if is_diagonal(self.fold_matrix):
            folds = np.diagonal(self.fold_matrix)
            self.pc_lattice = Lattice((self.structure.lattice.matrix.T * (1.0 / folds)).T.copy())
        else:
            raise NotImplementedError("non diagonal fold_matrix: %s" % str(self.fold_matrix))

        # Read fold2bloch output data.
        self.uf_kfrac_coords = self.reader.read_value("reduced_coordinates_of_unfolded_kpoints")
        self.uf_kpoints = KpointList(self.pc_lattice.reciprocal_lattice, self.uf_kfrac_coords)
        self.uf_nkpt = len(self.uf_kpoints)

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app(self.ebands.to_string(with_structure=False, title="Electronic Bands"))

        app("Direct lattice of the primitive cell:")
        to_s = lambda x: "%0.6f" % x
        app("abc   : " + " ".join([to_s(i).rjust(10) for i in self.pc_lattice.abc]))
        app("angles: " + " ".join([to_s(i).rjust(10) for i in self.pc_lattice.angles]))
        if is_diagonal(self.fold_matrix):
            app("Diagonal folding: %s" % (str(np.diagonal(self.fold_matrix))))
        else:
            app("Folding matrix: %s" % str(self.fold_matrix))

        if verbose:
            app(self.uf_kpoints.to_string(verbose=verbose, title="Unfolded k-points"))

        if verbose > 2:
            app(self.hdr.to_string(verbose=verbose, title="Abinit Header"))

        return "\n".join(lines)

    @property
    def ebands(self) -> ElectronBands:
        """|ElectronBands| object with folded band energies."""
        return self._ebands

    @property
    def structure(self) -> Structure:
        """|Structure| object defining the supercell."""
        return self.ebands.structure

    def close(self) -> None:
        """Close the file."""
        self.reader.close()

    @lazy_property
    def params(self) -> dict:
        """dict with parameters that might be subject to convergence studies."""
        od = self.get_ebands_params()
        return od

    @lazy_property
    def uf_eigens(self) -> np.ndarray:
        """[nsppol, nk_unfolded, nband] |numpy-array| with unfolded eigenvalues in eV."""
        # nctkarr_t("unfolded_eigenvalues", "dp", "max_number_of_states, nk_unfolded, number_of_spins")
        return self.reader.read_value("unfolded_eigenvalues") * units.Ha_to_eV

    @lazy_property
    def uf_weights(self) -> np.ndarray:
        """[nss, nk_unfolded, nband] array with spectral weights. nss = max(nspinor, nsppol)."""
        # nctkarr_t("spectral_weights", "dp", "max_number_of_states, nk_unfolded, nsppol_times_nspinor")
        return self.reader.read_value("spectral_weights")

    def get_spectral_functions(self, step=0.01, width=0.02):
        """
        Args:
            step: Energy step (eV) of the linear mesh.
            width: Standard deviation (eV) of the gaussian.

        Return:
            mesh, sfw, int_sfw
        """
        # Compute linear mesh.
        epad = 3.0 * width
        e_min = self.uf_eigens.min() - epad
        e_max = self.uf_eigens.max() + epad
        nw = int(1 + (e_max - e_min) / step)
        mesh, step = np.linspace(e_min, e_max, num=nw, endpoint=True, retstep=True)

        sfw = np.zeros((self.nss, self.uf_nkpt, nw))
        for spin in range(self.nss):
            for ik in range(self.uf_nkpt):
                for band in range(self.nband):
                    e = self.uf_eigens[spin, ik, band]
                    sfw[spin, ik] += self.uf_weights[spin, ik, band] * gaussian(mesh, width, center=e)

        try :
            from scipy.integrate import cumulative_trapezoid as cumtrapz
        except ImportError:
            from scipy.integrate import cumtrapz

        int_sfw = cumtrapz(sfw, x=mesh, initial=0.0)

        return dict2namedtuple(mesh=mesh, sfw=sfw, int_sfw=int_sfw)

    @add_fig_kwargs
    def plot_unfolded(self, kbounds, klabels, ylims=None, dist_tol=1e-12, verbose=0,
                      colormap="afmhot", facecolor="black", ax=None, fontsize=8, **kwargs) -> Figure:
        r"""
        Plot unfolded band structure with spectral weights.

        Args:
            klabels: dictionary whose keys are tuple with the reduced coordinates of the k-points.
                The values are the labels. e.g. ``klabels = {(0.0,0.0,0.0): "$\Gamma$", (0.5,0,0): "L"}``.
            ylims: Set the data limits for the y-axis. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If left (right) is None, default values are used
            dist_tol: A point is considered to be on the path if its distance from the line
                is less than dist_tol.
            verbose: Verbosity level.
            colormap: Have a look at the colormaps here and decide which one you like:
                http://matplotlib.sourceforge.net/examples/pylab_examples/show_colormaps.html
            facecolor:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: Legend and title fontsize.

        Returns: |matplotlib-Figure|
        """
        cart_bounds = [self.pc_lattice.reciprocal_lattice.get_cartesian_coords(c)
                       for c in np.reshape(kbounds, (-1, 3))]
        uf_cart = self.uf_kpoints.get_cart_coords()

        p = find_points_along_path(cart_bounds, uf_cart, dist_tol)
        if len(p.ikfound) == 0:
            cprint("Warning: find_points_along_path returned zero points along the path. Try to increase dist_tol.", "yellow")
            return None
        if verbose:
            uf_frac_coords = np.reshape([k.frac_coords for k in self.uf_kpoints], (-1, 3))
            fcoords = uf_frac_coords[p.ikfound]
            print("Found %d points along input k-path" % len(fcoords))
            print("k-points of path in reduced coordinates:")
            print(fcoords)

        fact = 8.0
        e0 = self.ebands.fermie
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        ax.set_facecolor(facecolor)

        xs = np.tile(p.dist_list, self.nband)
        marker_spin = {0: "^", 1: "v"} if self.nss == 2 else {0: "o"}
        for spin in range(self.nss):
            ys = self.uf_eigens[spin, p.ikfound, :] - e0
            ws = self.uf_weights[spin, p.ikfound, :]
            s = ax.scatter(xs, ys.T, s=fact * ws.T, c=ws.T,
                           marker=marker_spin[spin], label=None if self.nss == 1 else "spin %s" % spin,
                           linewidth=1, edgecolors='none', cmap=plt.get_cmap(colormap))
            plt.colorbar(s, ax=ax, orientation='vertical')

        ax.set_xticks(p.path_ticks, minor=False)
        ax.set_xticklabels(klabels, fontdict=None, minor=False, size=kwargs.pop("klabel_size", "large"))
        ax.grid(True)
        ax.set_ylabel('Energy (eV)')
        set_axlims(ax, ylims, "y")
        if self.nss == 2: ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        print("TODO: Add call to plot_unfolded")
        yield self.ebands.plot(show=False)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to nbpath. If `nbpath` is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("f2b = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(f2b)"),
            nbv.new_code_cell("f2b.ebands.plot();"),
            nbv.new_code_cell("#f2b.unfolded_kpoints.plot();"),
            nbv.new_code_cell(r"""\
# kbounds = [0, 1/2, 0, 0, 0, 0, 0, 0, 1/2]
# klabels = ["Y", "$\Gamma$", "X"]
# f2b.plot_unfolded(kbounds, klabels);"""),
        ])

        return self._write_nb_nbpath(nb, nbpath)



#!/usr/bin/env python
r"""
Gruneisen parameters
====================

This example shows how to analyze the Gruneisen parameters
computed by anaddb via finite difference.
See also v8/Input/t45.in
"""
import abipy.data as abidata
from abipy import abilab

# Open the file with abiopen
# Alternatively one can use the shell and `abiopen.py OUT_GRUNS.nc -nb`
# to open the file in a jupyter notebook.
ncfile = abilab.abiopen(abidata.ref_file("mg2si_GRUNS.nc"))

# Plot phonon DOSes computed by anaddb.
ncfile.plot_doses(title="DOSes available in the GRUNS file.")

# Plot phonon bands with markers
# sphinx_gallery_thumbnail_number = 2
ncfile.plot_phbands_with_gruns(title="Phonon bands with markers proportional to Gruneisen parameters + DOSes")

ncfile.plot_gruns_bs(title="Gruneisen along high-symmetry path.")

ncfile.plot_phbands_with_gruns(fill_with="gruns_fd",
                               title="Gruneisen parameters with finite differences.", with_doses=None)

ncfile.plot_gruns_scatter(units='cm-1',title="Scatter plot with Gruneisen parameters")

# Construct plotter object to analyze multiple phonon bands.
plotter = ncfile.get_plotter()
plotter.combiboxplot()

ncfile.close()



# coding: utf-8
"""Work subclasses related to effective mass calculations."""
from __future__ import annotations

import numpy as np
import os

from monty.json import jsanitize
from abipy.core.kpoints import build_segments
from abipy.abio.inputs import AbinitInput
from .nodes import Node
from .works import Work, PhononWork
from .flows import Flow


def _get_red_dirs_from_opts(red_dirs, cart_dirs, reciprocal_lattice) -> np.ndarray:
    """
    Helper function to compute the list of directions from user input. Return numpy array.
    """
    all_red_dirs = []
    if red_dirs is not None:
        all_red_dirs.extend(np.reshape(red_dirs, (-1, 3)))

    if cart_dirs is not None:
        frac_coords = reciprocal_lattice.get_fractional_coords(np.reshape(cart_dirs, (-1, 3)))
        all_red_dirs.extend(frac_coords)

    return np.reshape(all_red_dirs, (-1, 3))


class EffMassLineWork(Work):
    """
    Work for the computation of effective masses via finite differences along a k-line.
    Useful for cases such as NC+SOC where DFPT is not implemented or if one is interested
    in the non-parabolic behaviour of the energy dispersion.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: EffMassLineWork
    """

    @classmethod
    def from_scf_input(cls, scf_input: AbinitInput,
                       k0_list, step=0.01, npts=15,
                       red_dirs=[[1, 0, 0], [0, 1, 0], [0, 0, 1]], ndivsm=-20,
                       cart_dirs=None, den_node=None, manager=None) -> EffMassLineWork:
        """
        Build the Work from an |AbinitInput| representing a GS-SCF calculation.

        Args:
            scf_input: |AbinitInput| for GS-SCF used as template to generate the other inputs.
            k0_list: List with the reduced coordinates of the k-points where effective masses are wanted.
            step: Step for finite difference in Angstrom^-1
            npts: Number of points sampled around each k-point for each direction.
            red_dirs: List of reduced directions used to generate the segments passing through the k-point
            cart_dirs: List of Cartesian directions used to generate the segments passing through the k-point
            den_node: Path to the DEN file or Task object producing a DEN file.
                Can be used to avoid the initial SCF calculation if a DEN file is already available.
                If None, a GS calculation is performed.
            ndivsm: if > 0, it's the number of divisions for the smallest segment of the path (Abinit variable).
                if < 0, it's interpreted as the pymatgen `line_density` parameter in which the number of points
                in the segment is proportional to its length. Typical value: -20.
            manager: |TaskManager| instance. Use default if None.
        """
        if npts < 3:
            raise ValueError("Number of points: `%s` should be >= 3 for finite differences" % npts)

        # Define list of directions from user input.
        reciprocal_lattice = scf_input.structure.lattice.reciprocal_lattice
        all_red_dirs = _get_red_dirs_from_opts(red_dirs, cart_dirs, reciprocal_lattice)
        kpts = build_segments(k0_list, npts, step, all_red_dirs, reciprocal_lattice)

        # Now build NSCF input with explicit list of k-points.
        nscf_input = scf_input.make_nscf_kptopt0_input(kpts)

        new = cls(manager=manager)

        # Need to perform SCF run if DEN file is not available.
        scf_task = new.register_scf_task(scf_input) if den_node is None else Node.as_node(den_node)

        new.register_nscf_task(nscf_input, deps={scf_task: "DEN"})

        if ndivsm != 0:
            ebands_inp = scf_input.make_ebands_input(ndivsm=ndivsm)
            new.register_nscf_task(ebands_inp, deps={scf_task: "DEN"})

        return new


class EffMassDFPTWork(Work):
    """
    Work for the computation of effective masses with DFPT.
    Requires explicit list of k-points and range of bands.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: EffMassDFPTWork
    """

    @classmethod
    def from_scf_input(cls, scf_input: AbinitInput,
                       k0_list, effmass_bands_f90,
                       ngfft=None, den_node=None, manager=None) -> EffMassDFPTWork:
        """
        Build the Work from an |AbinitInput| representing a GS-SCF calculation.

        Args:
            scf_input: |AbinitInput| for GS-SCF used as template to generate the other inputs.
            k0_list: List with the reduced coordinates of the k-points where effective masses are wanted.
            effmass_bands_f90: (nkpt, 2) array with band range for effmas computation.
                WARNING: Assumes Fortran convention with indices starting from 1.
            ngfft: FFT divisions (3 integers). Used to enforce the same FFT mesh in the NSCF run as the one used for GS.
            den_node: Path to the DEN file or Task object producing a DEN file.
                Can be used to avoid the initial SCF calculation if a DEN file is already available.
                If None, a GS calculation is performed.
            manager: |TaskManager| instance. Use default if None.
        """
        multi = scf_input.make_dfpt_effmass_inputs(k0_list, effmass_bands_f90, ngfft=ngfft)
        nscf_input, effmass_input = multi[0], multi[1]

        new = cls(manager=manager)

        # Important: keep a reference to the Frohlich input that can be used to run EPH calculations if needed.
        new.frohlich_input = multi[2]

        # Need to perform SCF run if DEN file is not available
        scf_task = new.register_scf_task(scf_input) if den_node is None else Node.as_node(den_node)

        nscf_task = new.register_nscf_task(nscf_input, deps={scf_task: "DEN"})
        new.register_effmass_task(effmass_input, deps={scf_task: "DEN", nscf_task: "WFK"})

        return new


class EffMassAutoDFPTWork(Work):
    """
    Work for the automatic computation of effective masses with DFPT.
    Band extrema are automatically detected by performing a NSCF calculation
    along a high-symmetry k-path with ndivsm.
    Requires more computation that EffMassWork since input variables (kpoints and band range)
    are computed at runtime.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: EffMassAutoDFPTWork
    """

    @classmethod
    def from_scf_input(cls, scf_input: AbinitInput,
                       ndivsm=15, tolwfr=1e-20, den_node=None, manager=None) -> EffMassAutoDFPTWork:
        """
        Build the Work from an |AbinitInput| representing a GS-SCF calculation.

        Args:
            scf_input: |AbinitInput| for GS-SCF used as template to generate the other inputs.
            ndivsm: if > 0, it's the number of divisions for the smallest segment of the path (Abinit variable).
                if < 0, it's interpreted as the pymatgen `line_density` parameter in which the number of points
                in the segment is proportional to its length. Typical value: -20.
                This option is the recommended one if the k-path contains two high symmetry k-points that are very close
                as ndivsm > 0 may produce a very large number of wavevectors.
            tolwfr: Tolerance on residuals for NSCF calculation
            den_node: Path to the DEN file or Task object producing a DEN file.
                Can be used to avoid the initial SCF calculation if a DEN file is already available.
                If None, a GS calculation is performed.
            manager: |TaskManager| instance. Use default if None.
        """
        if scf_input.get("nsppol", 1) != 1:
            raise NotImplementedError("Magnetic semiconductors with nsppol = 2 are not implemented!")

        new = cls(manager=manager)
        # Keep a copy of the initial input.
        new.scf_input = scf_input.deepcopy()

        # Need SCF run to get DEN file.
        if den_node is None:
            new.den_node = new.register_scf_task(new.scf_input)
        else:
            new.den_node = Node.as_node(den_node)

        # Perform NSCF run along k-path that will be used to find band extrema.
        bands_input = scf_input.make_ebands_input(ndivsm=ndivsm, tolwfr=tolwfr)
        new.bands_task = new.register_nscf_task(bands_input, deps={new.den_node: "DEN"})

        return new

    def on_all_ok(self):
        """
        This method is called once the `Work` is completed i.e. when all tasks have reached status S_OK.
        Here, we read the band structure from GSR to find the position of the band edges and use these values
        to generate a new work for effective masses with DFPT.
        """
        with self.bands_task.open_gsr() as gsr:
            ebands = gsr.ebands
            # Warning: Assuming semiconductor with spin-unpolarized band energies.
            # At present, Abinit input variables do not take into account nsppol.
            ebands.set_fermie_to_vbm()
            # Find k0_list and effmass_bands_f90
            k0_list, effmass_bands_f90 = ebands.get_kpoints_and_band_range_for_edges()
            den_ngfft = gsr.reader.read_ngfft3()

        # Create the work for effective mass computation with DFPT and add it to the flow.
        # Keep a reference in generated_effmass_dfpt_work.
        work = EffMassDFPTWork.from_scf_input(self.scf_input, k0_list, effmass_bands_f90,
                                              ngfft=den_ngfft, den_node=self.den_node)

        self.generated_effmass_dfpt_work = work
        self.flow.register_work(work)
        self.flow.allocate()
        self.flow.build_and_pickle_dump()
        self.flow.finalized = False

        return super().on_all_ok()


class FrohlichZPRFlow(Flow):
    """
    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: FrohlichZPRFlow
    """

    @classmethod
    def from_scf_input(cls, workdir: str, scf_input: AbinitInput,
                       ddb_node=None, ndivsm=15, tolwfr=1e-20,
                       metadata=None, manager=None) -> FrohlichZPRFlow:
        """
        Build the Flow from an |AbinitInput| representing a GS-SCF calculation.
        Final results are stored in the "zprfrohl_results.json" in the outdata directory of the flow.

        Args:
            workdir: Working directory.
            scf_input: |AbinitInput| for GS-SCF used as template to generate the other inputs.
            ddb_node: Path to an external DDB file that is used to avoid the calculation of BECS/eps_inf and phonons.
                If None, a DFPT calculation is automatically performed by the flow.
            ndivsm: Number of divisions used to sample the smallest segment of the k-path.
            tolwfr: Tolerance on residuals for NSCF calculation
            manager: |TaskManager| instance. Use default if None.
            metadata: Dictionary with metadata addeded to the final JSON file.
        """
        new = cls(workdir=workdir, manager=manager)
        new.metadata = jsanitize(metadata) if metadata is not None else None

        # Build work for the automatic computation of effective masses.
        new.effmass_auto_work = EffMassAutoDFPTWork.from_scf_input(scf_input, ndivsm=ndivsm, tolwfr=tolwfr)
        new.register_work(new.effmass_auto_work)
        new.scf_task, new.ebands_kpath_task = new.effmass_auto_work[0], new.effmass_auto_work[1]

        if ddb_node is not None:
            new.ddb_node = Node.as_node(ddb_node)
            new.ddb_file_path_if_ddb_node = os.path.abspath(ddb_node)
        else:
            # Compute DDB with BECS and eps_inf.
            new.ddb_file_path_if_ddb_node = None
            ph_work = PhononWork.from_scf_task(new.scf_task, qpoints=[0, 0, 0],
                                               is_ngqpt=False, tolerance=None, with_becs=True,
                                               ddk_tolerance=None)
            new.register_work(ph_work)
            new.ddb_node = ph_work

        return new

    def on_all_ok(self):
        """
        This method is called when all the works in the flow have reached S_OK.
        This method shall return True if the calculation is completed or
        False if the execution should continue due to side-effects such as adding a new work to the flow.
        """
        if self.on_all_ok_num_calls > 0: return True
        self.on_all_ok_num_calls += 1

        work = Work()
        inp = self.effmass_auto_work.generated_effmass_dfpt_work.frohlich_input
        wfk_task = self.effmass_auto_work.generated_effmass_dfpt_work[0]
        self.effmass_task = self.effmass_auto_work.generated_effmass_dfpt_work[1]

        deps = {wfk_task: "WFK", self.ddb_node: "DDB", self.effmass_task: "EFMAS.nc"}
        self.frohl_task = work.register_eph_task(inp, deps=deps)
        self.register_work(work)

        self.allocate()
        self.build_and_pickle_dump()
        return False

    def finalize(self):
        """
        This method is called when the flow is completed.
        Here we write the final results in the "zprfrohl_results.json" file
        in the `outdata` directory of the flow
        """
        d = {}
        if self.metadata is not None: d.update({"metadata": self.metadata})

        # Add GS results.
        with self.scf_task.open_gsr() as gsr:
            d["gsr_scf_path"] = gsr.filepath
            d["pressure_GPa"] = float(gsr.pressure)
            d["max_force_eV_Ang"] = float(gsr.max_force)
            d["structure"] = gsr.structure

        # Add NSCF band structure.
        with self.ebands_kpath_task.open_gsr() as gsr:
            d["gsr_nscf_kpath"] = gsr.filepath
            gsr.ebands.set_fermie_to_vbm()
            d["ebands_kpath"] = gsr.ebands
            #d["ebands_info"] = gsr.ebands.get_dict4pandas(with_geo=False, with_spglib=False)

        # TODO
        # Extract results from run.abo
        #d["effmass_results"] = self.effmass_task.yaml_parse_results()
        #d["frohl_results"] = self.frohl_task.yaml_parse_results()

        # Add epsinf, e0, BECS, alpha and DDB as string.
        from abipy import abilab
        if self.ddb_file_path_if_ddb_node is not None:
            ddb_filepath = self.ddb_file_path_if_ddb_node
        else:
            ddb_filepath = self.ddb_node.outdir.path_in("out_DDB")

        with abilab.abiopen(ddb_filepath) as ddb:
            d["ddb_path"] = ddb.filepath
            d["ddb_string"] = ddb.get_string()
            epsinf, becs = ddb.anaget_epsinf_and_becs(chneut=1)
            gen = ddb.anaget_dielectric_tensor_generator(asr=2, chneut=1, dipdip=1)
            eps0 = gen.tensor_at_frequency(w=0.0)
            d["becs"] = becs
            d["epsinf_cart"] = epsinf
            # FIXME Complex is not supported by JSON.
            d["eps0_cart"] = eps0.real

        #print(d)
        abilab.mjson_write(d, self.outdir.path_in("zprfrohl_results.json"), indent=4)

        return super().finalize()



# coding: utf-8
from __future__ import print_function, division, unicode_literals, absolute_import

executable = "optic"

#from abimkdocs.variables import ValueWithUnit, MultipleValue, Range
from abipy.abio.abivar_database.variables import ValueWithUnit, MultipleValue, Range, ValueWithConditions
ValueWithConditions = dict
Variable=dict

variables = [
Variable(
    abivarname="broadening@optic",
    varset="optic",
    vartype="real",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval="1.d-3 Ha",
    mnemonics="BROADENING",
    added_in_version="before_v9",
    text=r"""
This parameter applies a broadening to the spectrum and is used to avoid
divergences in the sum-over-states approach.
The sum-over-states approach to the linear and nonlinear susceptibilities
inevitably include resonant denominators of the form
$(\omega - \omega_{nm})^{-1}$, see for example Eq. 46 of [[cite:Sharma2004]].
Numerically these denominators lead to infinities. In order to avoid them
one could do one of two things.
One could change the sum over k-points to integration, and then use the
linear tetrahedron method (see [[cite:Hughes1996]] for details).
Another way to get around the problem, as we will do in the present case,
is to avoid the singularities by adding a small imaginary contribution to the
denominator. This addition prevents the denominator from ever going to 0,
and acts as a broadening to the spectrum. The broadening should not be too
large, as this would wash out the features in the spectrum.
""",
),

Variable(
    abivarname="ddkfile@optic",
    varset="optic",
    vartype="string",
    topics=['Optic_basic'],
    dimensions="scalar",
    mnemonics="DDK FILE",
    commentdefault="no default",
    added_in_version="before_v9",
    text=r"""
This parameter specifies the name of the file containing the matrix elements of the
$d/dk$ operator in direction X, as the string ddkfile_X. This file should have been
produced by a preparatory Abinit run.
This file must not contain the first-order wavefunctions, and may be generated
using [[prtwf]] 3.

!!! important

    Make sure that the number of bands, spin channels, and k-points are the same in all the files.
""",
),

Variable(
    abivarname="domega@optic",
    varset="optic",
    vartype="real",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval="1.d-3 Ha",
    mnemonics="Delta OMEGA",
    added_in_version="before_v9",
    text=r"""
This parameter specifies the step size $\Delta\omega$ for the grid over which the
optic utility computes the susceptibilities. The maximum energy is set by the
companion parameter [[optic:maxomega]]. The susceptibilities are thus computed at
[[optic:maxomega]]/[[optic:domega]] energy points (zero excluded). In order
to capture more features, decrease the step size to get a finer energy
grid. In order to go to higher frequency, increase the maximum.
""",
),

Variable(
    abivarname="lin_comp@optic",
    varset="optic",
    vartype="integer",
    topics=['Optic_basic'],
    dimensions=[['num_lin_comp']],
    defaultval=0,
    mnemonics="LINear COMPonents",
    added_in_version="before_v9",
    text=r"""
This parameter specifies the directions of the [[optic:num_lin_comp]] requested components
of the dielectric tensor. The components are specified in
cartesian coordinates, where 1, 2, and 3 represent x, y, and z respectively. For
example, 11 represents the xx component, and 32 represents zy. There should be
[[optic:num_lin_comp]] entries. Note that these directions are denoted by
$a$ and $b$ in [[cite:Sharma2004]].
""",
),

Variable(
    abivarname="maxomega@optic",
    varset="optic",
    vartype="real",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval="1 Ha",
    mnemonics="MAXimum value of OMEGA",
    added_in_version="before_v9",
    text=r"""
This parameter specifies the maximum energy for the grid over which the
optic utility computes the susceptibilities. The grid step size is set by the
companion parameter [[optic:domega]]. The susceptibilities are thus computed at
[[optic:maxomega]]/[[optic:domega]] energy points (zero excluded). In order
to capture more features, decrease the step size to get a finer energy
grid. In order to go to higher frequency, increase the maximum.
""",
),

Variable(
    abivarname="nonlin_comp@optic",
    varset="optic",
    vartype="integer",
    topics=['Optic_basic'],
    dimensions=[['num_nonlin_comp']],
    defaultval=0,
    mnemonics="NON-LINear COMPonents",
    added_in_version="before_v9",
    text=r"""
This parameter specifies the directions of the [[optic:num_nonlin_comp]] requested components
of the second-order nonlinear dielectric tensor. The components are specified in
cartesian coordinates, where 1, 2, and 3 represent x, y, and z respectively. For
example, 111 represents the xxx component, and 321 represents zyx. There should be
[[optic:num_nonlin_comp]] entries. Note that these directions are denoted by
$a$ and $b$ in [[cite:Sharma2004]].
""",
),

Variable(
    abivarname="num_lin_comp@optic",
    varset="optic",
    vartype="integer",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval=0,
    mnemonics="NUMber of LINear COMPonents",
    added_in_version="before_v9",
    text=r"""
This parameter specifies how many components (out of 9 possible)
of the linear optical dielectric tensor to calculate.
Some of these may be either equal to each other, or zero, depending upon the
symmetry of the material (for details see [[cite:Draxl2006]]).
The directions of the requested components are specified by the parameter
[[optic:lin_comp]].
""",
),

Variable(
    abivarname="num_nonlin_comp@optic",
    varset="optic",
    vartype="integer",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval=0,
    mnemonics="NUMber of NON-LINear COMPonents",
    added_in_version="before_v9",
    text=r"""
This parameter specifies how many components (out of 27 possible)
of the second-order nonlinear optical tensor to calculate.
Some of these may be either equal to each other, or zero, depending upon the
symmetry of the material. The directions of the requested components are specified by the parameter [[optic:nonlin_comp]].
""",
),

Variable(
    abivarname="num_linel_comp@optic",
    varset="optic",
    vartype="integer",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval=0,
    mnemonics="NUMber of LINear ELetro-optic  COMPonents",
    added_in_version="before_v9",
    text=r"""
This parameter specifies how many components (out of 27 possible)
of the linear electro-optical susceptibility to calculate.
Some of these may be either equal to each other, or zero, depending upon the
symmetry of the material. The directions of the requested components are specified by the parameter [[optic:linel_comp]].
""",
),

Variable(
    abivarname="prtlincompmatrixelements@optic",
    varset="optic",
    vartype="integer",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval=0,
    mnemonics="PRinT the LINear COMPonent of the dielectric tensor's MATRIX ELEMENTS",
    added_in_version="v9.5",
    text=r"""
If set to 1, the matrix elements, the renormalized (but unshifted) Kohn-Sham eigenvalues,
the occupations and the kpts weights are all printed out in the ‘_OPTIC.nc’ file
generated by the optic tool. The matrix elements are the ones used to build the linear
components of the dielectric tensor. Useful for post processing of matrix elements or
rebuild the linear components of the dielectric tensor in an external script.
""",
),

Variable(
    abivarname="linel_comp@optic",
    varset="optic",
    vartype="integer",
    topics=['Optic_basic'],
    dimensions=[['num_linel_comp']],
    defaultval=0,
    mnemonics="LINear ELectro-optic COMPonents",
    added_in_version="before_v9",
    text=r"""
This parameter specifies the directions of the [[optic:num_linel_comp]] requested components
of the linear electro-optical susceptibility. The components are specified in
cartesian coordinates, where 1, 2, and 3 represent x, y, and z respectively. For
example, 111 represents the xxx component, and 321 represents zyx. There should be
[[optic:num_linel_comp]] entries.
""",
),

Variable(
    abivarname="scissor@optic",
    varset="optic",
    vartype="real",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval=0.0,
    mnemonics="SCISSOR operator",
    commentdefault="in Ha",
    added_in_version="before_v9",
    text=r"""
This parameter provides a fixed shift to all the conduction bands. As
LDA/GGA are known to underestimate the band-gap by a significant amount in
some cases, in order to obtain a reasonable optical spectrum and make a realistic
comparison with experiments one needs to correct for this [[cite:Levine1989]].
The scissors shift is normally chosen to be the difference between the experimental and
theoretical band-gap, and simply shifts the conduction bands. Alternatively, one may
determine the self energy using the [[tutorial:gw1|GW approach]], in which case
the opening of the gap due to the GW correction can be used as the scissor shift.
""",
),

Variable(
    abivarname="tolerance@optic",
    varset="optic",
    vartype="real",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval="1.d-3 Ha",
    mnemonics="TOLERANCE",
    added_in_version="before_v9",
    text=r"""
This parameter sets a scale for discarding small energy denominators.
When energy denominators are smaller than **tolerance**, the term is discarded from the sum.
See also [[optic:broadening]].
""",
),

Variable(
    abivarname="wfkfile@optic",
    varset="optic",
    vartype="string",
    topics=['Optic_basic'],
    dimensions="scalar",
    mnemonics="WaveFunction K FILE",
    commentdefault="no default",
    added_in_version="before_v9",
    text=r"""
This parameter specifies the name of the ground state wavefunction file, which
should have been produced in a preparatory Abinit run. It should include both
the valence and conduction states to be used in the optic calculation
(see [[help:optic]]).
""",
),

Variable(
    abivarname="nband_sum@optic",
    varset="optic",
    vartype="integer",
    topics=['Optic_basic'],
    dimensions="scalar",
    mnemonics="Number of BANDs in SUM",
    commentdefault="-1 i.e. use all bands found in external files. ",
    added_in_version="9.6.2",
    text=r"""
This variable specifies the number of bands included in the sum over states expressions.
If not given in input, all the bands found in the external files are used.

This option can be used to perform convergence studies with respect to the number of bands
for fixed BZ sampling without having to perform multiple calculations from scratch
just to change the number of states.
The idea is very simple: compute the WFK and the DDK files with a sufficiently large number of
bands then use nband_sum to change this parameter when running optic.

Obviously nband_sum cannot be greater than the number of bands stored on disk.
""",
),

]



#!/usr/bin/env python
r"""
Luminescent properties (1D-CCM)
===============================

This example shows how to post-process the results of a LumiWork
following a one-dimensional configuration coordinate model (1D-CCM).
Based on NV- center in diamond (64 atoms supercell)
"""

#%%
# Read the 4 points netcdf file produced by a LumiWork

from abipy.lumi.deltaSCF import DeltaSCF
import abipy.data as abidata
import abipy.abilab as abilab


NV_center = DeltaSCF.from_four_points_file([abidata.ref_file("relaxed_gs_out_GSR.nc"),
                                            abidata.ref_file("unrelaxed_ex_out_GSR.nc"),
                                            abidata.ref_file("relaxed_ex_out_GSR.nc"),
                                            abidata.ref_file("unrelaxed_gs_out_GSR.nc")])

# %%
# To draw the configuration coordinates diagram.

NV_center.draw_displaced_parabolas(scale_eff_freq=4);

#%%
# To plot the luminescence lineshape following the one effective phonon mode model at 0K

NV_center.plot_lineshape_1D_zero_temp(energy_range=[0.5,3],max_m=20,phonon_width=0.02,
               with_omega_cube='True',
               normalized='Sum');

#%%
# To get a panda dataframe with the main restults :
NV_center.get_dataframe()

#%%
# To plot the magnitude of the displacements/forces, from the N atom.

NV_center.plot_delta_R_distance(defect_symbol="N");
NV_center.plot_delta_F_distance(defect_symbol="N");

#%%
# To visualise displacements in 3D

NV_center.displacements_visu();

#%%
# To plot the four band structures associated to each point
# nscf_files are obtained in a LumiWork by activating the BS computations

nscf_files=[abidata.ref_file("relaxed_gs_nscf_GSR.nc"),
            abidata.ref_file("unrelaxed_ex_nscf_GSR.nc"),
            abidata.ref_file("relaxed_ex_nscf_GSR.nc"),
            abidata.ref_file("unrelaxed_gs_nscf_GSR.nc")]
NV_center.plot_four_BandStructures(nscf_files, ylims=[-4,4]);

# %%



# coding: utf-8
"""Density/potential files in netcdf/fortran format."""
from __future__ import annotations

import os
#import tempfile
#import numpy as np

from monty.string import marquee
from monty.termcolor import cprint
from monty.functools import lazy_property
from abipy.core.globals import get_workdir
from abipy.core.mixins import (AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter,
    AbinitFortranFile, CubeFile)
from abipy.core.structure import Structure
from abipy.core.fields import FieldReader
from abipy.electrons.ebands import ElectronBands, ElectronsReader
from abipy.abio.inputs import Cut3DInput
from abipy.flowtk import Cut3D



__all__ = [
    "DensityNcFile",
    "VhartreeNcFile",
    "VxcNcFile",
    "VhxcNcFile",
    "PotNcFile",
]


class Cut3dDenPotNcFile(AbinitNcFile, Has_Structure):
    """
    Netcdf file with structure and density/potential produced by CUT3d
    Unlike _NcFileWithField subclasses, this object does not contain an electronic band-structure
    and it's mainly used to convert from Fortran DEN/POT to netcdf.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: Cut3dDenPotNcFile
    """
    def __init__(self, filepath: str):
        super().__init__(filepath)
        self.reader = FieldReader(filepath)
        self.field = self.reader.read_field()

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.field.structure

    def close(self) -> None:
        """Close the file."""
        self.reader.close()

    @lazy_property
    def params(self) -> dict:
        """dict with parameters that might be subject to convergence studies."""
        return {}


class _DenPotNcReader(ElectronsReader, FieldReader):
    """Object used to read data from density/potential files in netcdf format."""


class _NcFileWithField(AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter):
    """
    Base class providing commong methods for netcdf files with density/potential
    """
    field_name = None

    @classmethod
    def from_file(cls, filepath: str):
        """Initialize the object from a Netcdf file"""
        return cls(filepath)

    def __init__(self, filepath: str):
        super().__init__(filepath)
        self.reader = _DenPotNcReader(filepath)

    @lazy_property
    def ebands(self) -> ElectronBands:
        """|ElectronBands| object."""
        return self.reader.read_ebands()

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.ebands.structure

    @lazy_property
    def xc(self):
        """:class:`XcFunc` object with info on the exchange-correlation functional."""
        return self.reader.read_abinit_xcfunc()

    @lazy_property
    def params(self) -> dict:
        """dict with parameters that might be subject to convergence studies."""
        od = self.get_ebands_params()
        return od

    @lazy_property
    def field(self):
        """
        The field object provided by the subclass.
        Methods of this base class should use self.field to implement
        logic common to the sub-classes.
        """
        return getattr(self, self.__class__.field_name)

    def close(self) -> None:
        """Close the file."""
        self.reader.close()

    def __str__(self) -> str:
        """String representation."""
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app(self.ebands.to_string(with_structure=False, title="Electronic Bands"))
        app("XC functional: %s" % str(self.xc))

        # Add info on the field
        app(marquee(self.field.__class__.__name__, mark="="))
        app(str(self.field))

        if verbose > 1:
            app("")
            app(self.hdr.to_string(verbose=verbose, title="Abinit Header"))

        return "\n".join(lines)

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        yield self.structure.plot(show=False)
        yield self.ebands.plot(show=False)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("ncfile = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(ncfile)"),
            nbv.new_code_cell("ncfile.ebands.kpoints.plot();"),
            nbv.new_code_cell("ncfile.ebands.plot();"),
            nbv.new_code_cell("ncfile.ebands.get_edos().plot();"),
            nbv.new_code_cell("#cube = ncfile.write_cube(filename=None)"),
            nbv.new_code_cell("#xsf_path = ncfile.write_xsf(filename=None"),
            nbv.new_code_cell("#chgcar = ncfile.write_chgcar(filename=None"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class DensityNcFile(_NcFileWithField):
    """
    Netcdf File containing the electronic density.

    Usage example:

    .. code-block:: python

        with DensityNcFile("foo_DEN.nc") as ncfile:
            ncfile.density
            ncfile.ebands.plot()

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: DensityNcFile
    """
    field_name = "density"

    @lazy_property
    def density(self):
        """Density object."""
        return self.reader.read_density()

    def to_string(self, verbose=0) -> str:
        """String representation."""
        s = super().to_string(verbose=verbose)

        # Add density related stuff.
        lines = [" "]
        app = lines.append
        app("Integrated electronic and magnetization densities in atomic spheres:")
        df = self.density.integrate_in_spheres(rcut_symbol=None, out=False)
        app(str(df))
        app("Total magnetization from unit cell integration: %s" % str(self.density.magnetization))

        return s + "\n".join(lines)

    def write_chgcar(self, filename=None):
        """
        Write density in CHGCAR format. Return :class:`ChgCar` instance.
        """
        if filename is None:
            filename = self.basename.replace(".nc", "_CHGCAR")
            cprint("Writing density in CHGCAR format to file: %s" % filename, "yellow")

        return self.density.to_chgcar(filename=filename)

    def write_xsf(self, filename=None):
        """
        Write density in XSF format (xcrysden_)
        """
        if filename is None:
            filename = self.basename.replace(".nc", ".xsf")
            cprint("Writing density in XSF format to file: %s" % filename, "yellow")

        return self.density.export(filename)

    def write_cube(self, filename=None, spin="total"):
        """Write density in CUBE format."""
        if filename is None:
            filename = self.basename.replace(".nc", ".cube")
            cprint("Writing density in CUBE format to file: %s" % filename, "yellow")

        return self.density.export_to_cube(filename, spin=spin)


class VhartreeNcFile(_NcFileWithField):
    """
    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: VhartreeNcFile
    """
    field_name = "vh"

    @lazy_property
    def vh(self):
        """Hartree potential."""
        return self.reader.read_vh()


class VxcNcFile(_NcFileWithField):
    """
    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: VxcNcFile
    """
    field_name = "vxc"

    @lazy_property
    def vxc(self):
        """XC potential."""
        return self.reader.read_vxc()


class VhxcNcFile(_NcFileWithField):
    """
    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: VhxcNcFile
    """
    field_name = "vhxc"

    @lazy_property
    def vhxc(self):
        """Hartree + XC potential."""
        return self.reader.read_vhxc()


class PotNcFile(_NcFileWithField):
    """
    POT file produced by the GS part containing the total KS potential in real space.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: PotNcFile
    """
    field_name = "vks"

    @lazy_property
    def vks(self):
        """Hartree + XC potential + sum of local pseudo-potential terms."""
        return self.reader.read_vks()


class DfptPotNcFile(_NcFileWithField):
    """
    POT file produced by the DFPT part containing the
    first order derivative of the KS potential in real space.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: DfptPotNcFile
    """
    field_name = "first_order_potential"

    #@lazy_property
    #def vk1s(self):
    #    """
    #    First order derivative of KS potential.
    #    Includes derivative of Hartree + XC potential + sum of local pseudo-potential terms.
    #    except derivative of non-local part.
    #    """
    #    return self.reader.read_vks1()


class DensityFortranFile(AbinitFortranFile):
    """
    Class representing the _DEN fortran file containing density.
    Provides methods to run Cut3D and convert data into different formats.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: DensityFortranFile
    """

    def _convert(self, cut3d_input, workdir=None):
        """
        Internal function to run a conversion using cut3d.
        """
        workdir = get_workdir(workdir)
        outfile, converted_file = Cut3D().cut3d(cut3d_input, workdir)

        return converted_file

    def get_cube(self, out_filepath, workdir=None) -> CubeFile:
        """
        Runs cut3d to convert the density to the cube format

        Args:
            out_filepath: path to the file that should be produced by cut3D, if required. At this stage it would be
                safer to use just the file name, as using an absolute or relative path may fail depending on the compiler.
            workdir: directory where cut3d is executed.

        Returns:
            (CubeFile) the converted file as a CubeFile object.
        """
        return CubeFile(self._convert(cut3d_input=Cut3DInput.den_to_cube(self.filepath, out_filepath),
                                      workdir=workdir))

    def get_xsf(self, out_filepath, shift=None, workdir=None):
        """
        Runs cut3d to convert the density to the xsf format

        Args:
            out_filepath: path to the file that should be produced by cut3D, if required. At this stage it would be
                safer to use just the file name, as using an absolute or relative path may fail depending on
                the compiler.
            shift: a list of three integers defining the shift along the x, y, z axis. None if no shift is required.
            workdir: directory where cut3d is executed.

        Returns:
            (string) path to the converted file.
        """
        return self._convert(cut3d_input=Cut3DInput.den_to_xsf(self.filepath,
                             output_filepath=out_filepath, shift=shift), workdir=workdir)

    def get_tecplot(self, out_filepath, workdir=None):
        """
        Runs cut3d to convert the density to the tecplot format

        Args:
            out_filepath: path to the file that should be produced by cut3D, if required. At this stage it would be
                safer to use just the file name, as using an absolute or relative path may fail depending on
                the compiler.
            workdir: directory where cut3d is executed.

        Returns:
            (string) path to the converted file.
        """
        return self._convert(cut3d_input=Cut3DInput.den_to_tecplot(self.filepath, out_filepath),
                             workdir=workdir)

    def get_molekel(self, out_filepath, workdir=None):
        """
        Runs cut3d to convert the density to the molekel format

        Args:
            out_filepath: path to the file that should be produced by cut3D, if required. At this stage it would be
                safer to use just the file name, as using an absolute or relative path may fail depending on
                the compiler.
            workdir: directory where cut3d is executed.

        Returns:
            (string) path to the converted file.
        """
        return self._convert(cut3d_input=Cut3DInput.den_to_molekel(self.filepath, out_filepath),
                             workdir=workdir)

    def get_3d_indexed(self, out_filepath, workdir=None):
        """
        Runs cut3d to convert the density to the 3D indexed format

        Args:
            out_filepath: path to the file that should be produced by cut3D, if required. At this stage it would be
                safer to use just the file name, as using an absolute or relative path may fail depending on
                the compiler.
            workdir: directory where cut3d is executed.

        Returns:
            (string) path to the converted file.
        """
        return self._convert(cut3d_input=Cut3DInput.den_to_3d_indexed(self.filepath, out_filepath),
                             workdir=workdir)

    def get_3d_formatted(self, out_filepath, workdir=None):
        """
        Runs cut3d to convert the density to the 3D formatted format

        Args:
            out_filepath: path to the file that should be produced by cut3D, if required. At this stage it would be
                safer to use just the file name, as using an absolute or relative path may fail depending on
                the compiler.
            workdir: directory where cut3d is executed.

        Returns:
            (string) path to the converted file.
        """
        return self._convert(cut3d_input=Cut3DInput.den_to_3d_indexed(self.filepath, out_filepath),
                             workdir=workdir)

    def get_hirshfeld(self, structure, all_el_dens_paths=None, fhi_all_el_path=None, workdir=None):
        """
        Runs cut3d to get the Hirshfeld charges. Requires all-electron density files, so at least one source
        should be specified (all_el_dens_paths or fhi_all_el_path)

        Args:
            structure: a Structure object representing the structure used to generate the density
            all_el_dens_paths: a list of paths to the all-electron density files correspinding to the elements defined
                in the abinit input.
            fhi_all_el_path: path to the folder containing the fhi all-electron density files that will be used
                to automatically determine the path of the required densities.
            workdir: directory where cut3d is executed.

        Returns:
            (HirshfeldCharges) the calculated Hirshfeld charges.
        """
        if all_el_dens_paths is None and fhi_all_el_path is None:
            raise ValueError("At least one source of all electron densities should be provided.")
        if all_el_dens_paths is not None and fhi_all_el_path is not None:
            raise ValueError("all_el_dens_paths and fhi_all_el_path are mutually exclusive.")

        if all_el_dens_paths is not None:
            cut3d_input = Cut3DInput.hirshfeld(self.filepath, all_el_dens_paths)
        else:
            cut3d_input = Cut3DInput.hirshfeld_from_fhi_path(self.filepath, structure, fhi_all_el_path)

        workdir = get_workdir(workdir)

        cut3d = Cut3D()
        outfile, converted_file = cut3d.cut3d(cut3d_input, workdir)

        from abipy.electrons.charges import HirshfeldCharges
        return HirshfeldCharges.from_cut3d_outfile(structure=structure, filepath=cut3d.stdout_fname)

    def get_density(self, workdir=None):
        """
        Invoke cut3d to produce a netcdf file with the density, read the file and return Density object.

        Args:
            workdir: directory in which cut3d is executed.
        """
        workdir = get_workdir(workdir)
        output_filepath = os.path.join(workdir, "field_CUT3DDENPOT.nc")
        # FIXME Converters with nspden > 1 won't work since cut3d asks for the ispden index.
        cut3d_input = Cut3DInput(infile_path=self.filepath, output_filepath=output_filepath,
                                 options=[15, output_filepath, 0, 0])

        outfile, _ = Cut3D().cut3d(cut3d_input, workdir)
        with Cut3dDenPotNcFile(output_filepath) as nc:
            assert nc.field.is_density_like and nc.field.netcdf_name == "density"
            return nc.field



# coding: utf-8
"""
Base class for parsers
"""
from __future__ import annotations

import abc
import os


class ParserError(Exception):
    """Exceptions raised by Parser objects."""


class BaseParser(metaclass=abc.ABCMeta):
    """
    Abstract class defining the interface that must be provided
    by the parsers used to extract results from the output file of
    a pseudopotential generator a.k.a. ppgen

    Attributes:

        errors: List of strings with errors reported by the pp generator
        warnings: List of strings with the warnings reported by the pp generator.
    """

    Error = ParserError

    def __init__(self, filepath: str) -> None:
        self.filepath = os.path.abspath(filepath)
        self.workdir = os.path.abspath(os.path.dirname(filepath))
        self.run_completed = False
        self._errors = []
        self._warnings = []

    @property
    def errors(self) -> list[str]:
        """
        List of strings with possible errors reported by the generator at run-time.
        """
        return self._errors

    @property
    def warnings(self) -> list[str]:
        """
        List of strings with possible errors reported by the generator at run-time.
        """
        return self._warnings

    @abc.abstractmethod
    def get_results(self) -> dict:
        """
        Return the most important results in a dictionary.
        """

    @abc.abstractmethod
    def get_input_str(self) -> str:
        """Returns a string with the input file."""



# coding: utf-8
"""
Arpese Plotter (still under development)
"""
import numpy as np

from scipy.interpolate import UnivariateSpline
from monty.collections import dict2namedtuple
from abipy.core.mixins import Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.electrons import ElectronBands
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt, get_ax3d_fig_plt, get_axarray_fig_plt #set_axlims,


class ArpesPlotter(Has_Structure, Has_ElectronBands, NotebookWriter):
    """
    Usage example:

    .. code-block:: python

        with abilab.abiopen("foo_ABIWAN.nc") as abiwan:
            print(abiwan)

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: ArpesPlotter
    """
    @classmethod
    def model_from_ebands(cls, ebands, tmesh=(0, 300, 600), poorman_polaron=False):
        ebands = ElectronBands.as_ebands(ebands)

        ntemp = len(tmesh)
        nwr = 1000
        wr_step = 0.01

        aw = np.empty((ebands.nsppol, ebands.nkpt, ebands.mband, ntemp, nwr))
        aw_meshes = np.empty((ebands.nsppol, ebands.nkpt, ebands.mband, nwr))
        #aw: [nwr, ntemp, max_nbcalc, nkcalc, nsppol] array
        #aw_meshes: [max_nbcalc, nkcalc, nsppol] array with energy mesh in eV
        from abipy.tools.numtools import lorentzian
        try:
            from scipy.integrate import cumulative_trapezoid as cumtrapz
        except ImportError:
            from scipy.integrate import cumtrapz
        for spin in ebands.spins:
            for ik, kpt in enumerate(ebands.kpoints):
                for band in range(ebands.nband_sk[spin, ik]):
                    e0 = ebands.eigens[spin, ik, band]
                    emin = e0 - wr_step * (nwr // 2)
                    emax = e0 + wr_step * (nwr // 2)
                    emesh = np.linspace(emin, emax, num=nwr)
                    aw_meshes[spin, ik, band] = emesh
                    # Naive model: lorentzian centered on KS energy with T-dep broadening
                    for itemp, temp in enumerate(tmesh):
                        width = 0.2 + (temp / 300) * 0.2
                        avals = lorentzian(emesh, width=width, center=e0, height=None)
                        if poorman_polaron:
                            if band in (1, 2, 3) and kpt.norm < 0.3:
                                avals += 1.1 * lorentzian(emesh, width=0.1 * width, center=e0 - 0.4, height=None)
                            avals /= cumtrapz(avals, x=emesh)[-1]
                        aw[spin, ik, band, itemp] = avals

        return cls(ebands, aw, aw_meshes, tmesh)

    def __init__(self, ebands, aw, aw_meshes, tmesh):
        """
        Args:
            ebands: |ElectronBands| object
            aw: [nwr, ntemp, max_nbcalc, nkcalc, nsppol] array
            aw_meshes: [max_nbcalc, nkcalc, nsppol] array with energy mesh in eV
            tmesh: Temperature mesh in Kelvin.

        .. note::

            - Each (spin, k-point, band) can have different frequency meshes
              (usually centered on the initial KS energy).
              The plotter will spline the data.

            - The order of the k-points in ebands and aw should be the same.

            - The treatment of bands if complicated by the fact that we can have
              different nband(k) whose indices are not necessarily aligned.
              Consider, for example. what happes if we use symsigma or arbitrary bdgw.
              Use MaskedArray or metadata with start, stop?
        """
        self._ebands = ebands
        self.aw = aw
        self.aw_meshes = aw_meshes
        self.tmesh = tmesh
        self.ntemp = len(tmesh)
        #assert

        # Options passed to UnivariateSpline
        self.ext, self.k, self.s = "zeros", 3, 0

    @property
    def structure(self):
        """|Structure| object."""
        return self.ebands.structure

    @property
    def ebands(self):
        """|ElectronBands| object."""
        return self._ebands

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose=0):
        """String representation with verbosity level `verbose`."""
        lines = []; app = lines.append
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app(self.ebands.to_string(with_structure=False, verbose=verbose, title="Electronic Bands"))

        #if verbose > 1:
        return "\n".join(lines)

    def with_points_along_path(self, frac_bounds=None, knames=None, dist_tol=1e-12):
        """
        Args:
            frac_bounds: [M, 3] array  with the vertexes of the k-path in reduced coordinates.
                If None, the k-path is automatically selected from the structure.
            knames: List of strings with the k-point labels defining the k-path. It has precedence over frac_bounds.
            dist_tol: A point is considered to be on the path if its distance from the line
                is less than dist_tol.
        """
        r = self.ebands.with_points_along_path(frac_bounds=frac_bounds, knames=knames, dist_tol=dist_tol)
        # Transfer data using r.ik_new2prev table.
        return self.__class__(r.ebands,
                              aw=self.aw[:, :, :, r.ik_new2prev, :].copy(),
                              aw_meshes=self.aw_meshes[:, r.ik_new2prev, :].copy(),
                              tmesh=self.tmesh)

    #def interpolate(self):
    #    new_ebands = self.ebands.interpolate(lpratio=5, vertices_names=None, line_density=20,
    #                        kmesh=None, is_shift=None, filter_params=None, verbose=0)

    #    # Build interpolator.
    #    #from abipy.core.skw import SkwInterpolator
    #    #my_kcoords = [k.frac_coords for k in self.kpoints]
    #    #cell = (self.structure.lattice.matrix, self.structure.frac_coords,
    #    #        self.structure.atomic_numbers)

    #    #skw = SkwInterpolator(lpratio, my_kcoords, self.eigens, self.fermie, self.nelect,
    #    #                      cell, fm_symrel, self.has_timrev,
    #    #                      filter_params=filter_params, verbose=verbose)

    #    # Interpolate energies.
    #    #eigens_kpath = skw.interp_kpts(kfrac_coords).eigens

    #    return self.__class__(new_ebands, new_aw, aw_meshes, self.tmesh)

    def get_emesh_eminmax(self, estep):
        """Compute linear mesh covering entire energy range."""
        emin = self.ebands.enemin()
        emin -= 0.1 * abs(emin)
        emax = self.ebands.enemax()
        emax += 0.1 * abs(emax)
        return np.arange(emin, emax, estep), emin, emax

    def get_data_nmtuple(self, itemp, estep, spins=None):
        nkpt = self.ebands.nkpt
        spins = range(self.ebands.nsppol) if spins is None else spins

        emesh, emin, emax = self.get_emesh_eminmax(estep)
        nene = len(emesh)
        #print("nkpt", nkpt, "nene", nene)
        data = np.zeros((nkpt, nene))

        # aw: [nwr, ntemp, max_nbcalc, nkcalc, nsppol] array
        for spin in spins:
            for ik in range(nkpt):
                for band in range(self.ebands.nband_sk[spin, ik]):
                    w = self.aw_meshes[spin, ik, band]
                    aw = self.aw[spin, ik, band, itemp]
                    data[ik] += UnivariateSpline(w, aw, k=self.k, s=self.s, ext=self.ext)(emesh)

        return dict2namedtuple(data=data, emesh=emesh, emin=emin, emax=emax, spins=spins, nkpt=nkpt)

    def get_atw(self, wmesh, spin, ikpt, band_inds, temp_inds):
        ntemp, nene = len(temp_inds), len(wmesh)
        atw = np.zeros((ntemp, nene))
        for band in range(self.ebands.nband_sk[spin, ikpt]):
            if band_inds is not None and band not in band_inds: continue
            w = self.aw_meshes[spin, ikpt, band]
            for it, itemp in enumerate(temp_inds):
                aw = self.aw[spin, ikpt, band, itemp]
                atw[it] += UnivariateSpline(w, aw, k=self.k, s=self.s, ext=self.ext)(wmesh)

        return atw

    @add_fig_kwargs
    def plot_ekmap_temps(self, temp_inds=None, spins=None, estep=0.02, with_colorbar=True,
                        ylims=None, fontsize=8, **kwargs):
        """
        Plot (k, e) color maps for different temperatures.

        Args:
            fontsize (int): fontsize for titles and legend

        Return: |matplotlib-Figure|
        """
        temp_inds = range(self.ntemp) if temp_inds is None else temp_inds
        # Build plot grid.
        num_plots, ncols, nrows = len(temp_inds), 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=True, squeeze=False)
        ax_list = ax_list.ravel()

        # Don't show the last ax if numeb is odd.
        if num_plots % ncols != 0: ax_list[-1].axis("off")

        for itemp, ax in zip(temp_inds, ax_list):
            self.plot_ekmap_itemp(itemp=itemp, spins=spins, estep=estep, ax=ax, ylims=ylims,
                    with_colorbar=with_colorbar, show=False, **kwargs)
            ax.set_title("T = %.1f K" % self.tmesh[itemp], fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_ekmap_itemp(self, itemp=0, spins=None, estep=0.02, ax=None, ylims=None, with_colorbar=True, **kwargs):
        """
        Plot (k, e) color map for given temperature.

        Args:
            e0: Option used to define the zero of energy in the band structure plot. Possible values:
                - ``fermie``: shift all eigenvalues to have zero energy at the Fermi energy (``self.fermie``).
                - Number e.g ``e0 = 0.5``: shift all eigenvalues to have zero energy at 0.5 eV
                - None: Don't shift energies, equivalent to ``e0 = 0``.
            spins: Selects the spin to be plotted, None if all spins.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            ylims: Set the data limits for the y-axis. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If left (right) is None, default values are used
            with_colorbar: True to add color bar.
            kwargs: options passed to ``ax.imshow``.

        Return: |matplotlib-Figure|
        """
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        a = self.get_data_nmtuple(itemp, estep, spins=spins)

        cmap = "jet"
        img = ax.imshow(a.data.T, origin="lower", extent=[0, a.nkpt, a.emin, a.emax],
                  cmap=cmap,
                  interpolation="bilinear",
                  #interpolation="spline36",
                  #interpolation="bicubic",
                  #vmin=0, vmax=np.abs(data).max()
                  )
        self.ebands.plot(ax=ax, e0=0, show=False, color="w", lw=1, ls="--")

        #ax.set_zlabel(r"$A(\omega)$")
        #self.ebands.plot(ax=ax, e0=0, color="r", lw=1)
        #ax.imshow(data, cmap=None, norm=None, aspect=None, interpolation=None, alpha=None, vmin=None, vmax=None,
        #       origin=None, extent=None, shape=None, filternorm=1, filterrad=4.0, imlim=None, resample=None,
        #       url=None, hold=None, data=None, **kwargs)

        if with_colorbar:
            # Make a color bar
            #plt.colorbar(img, cmap=cmap)
            # https://stackoverflow.com/questions/13310594/positioning-the-colorbar
            from mpl_toolkits.axes_grid1 import make_axes_locatable
            divider = make_axes_locatable(ax)
            #cax = divider.new_vertical(size="5%", pad=0.1, pack_start=True)
            #cax = divider.new_horizontal(size="5%", pad=0.1, pack_start=True)

            # create an axes on the right side of ax. The width of cax will be 5%
            # of ax and the padding between cax and ax will be fixed at 0.05 inch.
            # https://matplotlib.org/2.0.2/mpl_toolkits/axes_grid/users/overview.html#axesdivider
            cax = divider.append_axes("right", size="5%", pad=0.05)

            #fig.add_axes(cax)
            #fig.colorbar(img, cax=cax, ax=ax, orientation="horizontal")
            fig.colorbar(img, cax=cax, ax=ax)

        return fig

    @add_fig_kwargs
    def plot_ak_vs_temp(self, temp_inds=None, spins=None, band_inds=None, kpt_inds=None,
                        apad=1.0, estep=0.02, colormap="jet", fontsize=8, **kwargs):
        """

        Args:
            temp_inds:
            spins:
            band_inds:
            kpt_inds:
            estep:
            colormap
            fontsize (int): fontsize for titles and legend

        Return: |matplotlib-Figure|
        """
        temp_inds = range(self.ntemp) if temp_inds is None else temp_inds
        ntemp = len(temp_inds)
        spins = range(self.ebands.nsppol) if spins is None else spins
        kpt_inds = range(self.ebands.nkpt) if kpt_inds is None else kpt_inds
        nkpt = len(kpt_inds)

        xs, emin, emax = self.get_emesh_eminmax(estep)
        nene = len(xs)

        num_plots, ncols, nrows = nkpt, 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        # Build plot grid.
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=True, squeeze=False)
        ax_list = np.array(ax_list).ravel()
        cmap = plt.get_cmap(colormap)

        for isp, spin in enumerate(spins):
            spin_sign = +1 if spin == 0 else -1
            for ik, (ikpt, ax) in enumerate(zip(kpt_inds, ax_list)):
                ax.grid(True)
                atw = self.get_atw(xs, spin, ikpt, band_inds, temp_inds)
                for it, itemp in enumerate(temp_inds):
                    ys = spin_sign * atw[it] + (it * apad)
                    ax.plot(xs, ys, lw=2, alpha=0.8, color=cmap(float(it) / ntemp),
                            label="T = %.1f K" % self.tmesh[itemp] if (ik, isp) == (0, 0) else None)

                if spin == 0:
                    kpt = self.ebands.kpoints[ikpt]
                    ax.set_title("k:%s" % (repr(kpt)), fontsize=fontsize)

                if (ik, isp) == (0, 0):
                    ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    #@add_fig_kwargs
    #def plot_ak(self, temp_inds=None, spins=None, band_inds=None, kpt_inds=None,
    #           apad=1.0, estep=0.02, colormap="jet", fontsize=8, **kwargs):
    #    """

    #    Args:
    #        temp_inds:
    #        spins:
    #        band_inds:
    #        kpt_inds:
    #        estep:
    #        colormap
    #        fontsize (int): fontsize for titles and legend

    #    Return: |matplotlib-Figure|
    #    """
    #    temp_inds = range(self.ntemp) if temp_inds is None else temp_inds
    #    ntemp = len(temp_inds)
    #    spins = range(self.ebands.nsppol) if spins is None else spins
    #    kpt_inds = range(self.ebands.nkpt) if kpt_inds is None else kpt_inds
    #    nkpt = len(kpt_inds)

    #    xs, emin, emax = self.get_emesh_eminmax(estep)
    #    nene = len(xs)

    #    num_plots, ncols, nrows = nkpt, 1, 1
    #    if num_plots > 1:
    #        ncols = 2
    #        nrows = (num_plots // ncols) + (num_plots % ncols)

    #    # Build plot grid.
    #    ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
    #                                            sharex=True, sharey=True, squeeze=False)
    #    ax_list = np.array(ax_list).ravel()
    #    cmap = plt.get_cmap(colormap)

    #    for isp, spin in enumerate(spins):
    #        spin_sign = +1 if spin == 0 else -1
    #        for ik, (ikpt, ax) in enumerate(zip(kpt_inds, ax_list)):
    #            ax.grid(True)
    #            atw = self.get_atw(xs, spin, ikpt, band_inds, temp_inds)
    #            for it, itemp in enumerate(temp_inds):
    #                ys = spin_sign * atw[it] + (it * apad)
    #                ax.plot(xs, ys, lw=2, alpha=0.8, color=cmap(float(it) / ntemp),
    #                        label = "T = %.1f K" % self.tmesh[itemp] if (ik, isp) == (0, 0) else None)

    #            if spin == 0:
    #                kpt = self.ebands.kpoints[ikpt]
    #                ax.set_title("k:%s" % (repr(kpt)), fontsize=fontsize)

    #            if (ik, isp) == (0, 0):
    #                ax.legend(loc="best", fontsize=fontsize, shadow=True)

    #    return fig

    @add_fig_kwargs
    def plot_3dlines(self, itemp=0, estep=0.02, spins=None, band_inds=None, ax=None, **kwargs):
        ax, fig, plt = get_ax3d_fig_plt(ax=ax)

        xs, emin, emax = self.get_emesh_eminmax(estep)
        nene = len(xs)
        nkpt = self.ebands.nkpt
        cmap = plt.get_cmap("jet")
        # TODO: Reshift everything?
        if band_inds is not None:
            band_inds = set(band_inds)

        # aw: [nwr, ntemp, max_nbcalc, nkcalc, nsppol] array
        spins = range(self.ebands.nsppol) if spins is None else spins
        for spin in spins:
            for ik in range(nkpt):
                ys = np.ones(nene) * ik
                zs = np.zeros(nene)
                for band in range(self.ebands.nband_sk[spin, ik]):
                    if band_inds is not None and band not in band_inds: continue
                    w = self.aw_meshes[spin, ik, band]
                    aw = self.aw[spin, ik, band, itemp]
                    zs += UnivariateSpline(w, aw, k=self.k, s=self.s, ext=self.ext)(xs)

                ax.plot(ys, xs, zs, color="k", lw=1, alpha=0.8) #cmap(float(ik) / nkpt))

                # Code to convert data in 3D polygons
                # See https://stackoverflow.com/questions/33641551/vertically-fill-3d-matplotlib-plot
                #h = 0.0
                #from mpl_toolkits.mplot3d.art3d import Poly3DCollection
                #xs = xs.copy()
                #ys = xs.copy()
                #zs = zs.copy()
                #v = []
                #for k in range(0, len(xs) - 1):
                #    x = [xs[k], xs[k+1], xs[k+1], xs[k]]
                #    y = [ys[k], ys[k+1], ys[k+1], ys[k]]
                #    z = [zs[k], zs[k+1],       h,     h]
                #    v.append(list(zip(x, y, z)))
                #poly3dCollection = Poly3DCollection(v)
                #ax.add_collection3d(poly3dCollection)

        ax.set_zlabel(r"$A(\omega)$")
        self.ebands.plot(ax=ax, e0=0, color="r", lw=1)

        return fig

    @add_fig_kwargs
    def plot_surface(self, itemp=0, estep=0.02, spins=None, ax=None, **kwargs):
        ax, fig, plt = get_ax3d_fig_plt(ax=ax)

        xs, emin, emax = self.get_emesh_eminmax(estep)
        nene = len(xs)
        nkpt = self.ebands.nkpt
        cmap = plt.get_cmap("jet")

        # aw: [nwr, ntemp, max_nbcalc, nkcalc, nsppol] array
        spins = range(self.ebands.nsppol) if spins is None else spins
        spin = 0
        zs = np.zeros((nkpt, nene))
        ys = np.arange(nkpt)
        for ik in range(nkpt):
            for band in range(self.ebands.nband_sk[spin, ik]):
                w = self.aw_meshes[spin, ik, band]
                aw = self.aw[spin, ik, band, itemp]
                zs[ik] += UnivariateSpline(w, aw, k=self.k, s=self.s, ext=self.ext)(xs)

        # Plot the surface.
        xs, ys = np.meshgrid(xs, ys)

        #surf = ax.plot_surface(ys, xs, zs,
        #                       rstride=1, cstride=1,
        #                       linewidth=0, antialiased=True,
        #                       #cmap=cmap,
        #                       )
        #cb = fig.colorbar(surf, shrink=0.5)

        ax.plot_wireframe(ys, xs, zs, rstride=1, cstride=5)

        ax.set_zlabel(r"$A(\omega)$")
        self.ebands.plot(ax=ax, e0=0, color="r", lw=1)

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield None
        # TODO
        #yield self.combiplot(show=False)
        #yield self.gridplot(show=False)

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to nbpath. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        # Use pickle files for data persistence.
        tmpfile = self.pickle_dump()

        # TODO
        nb.cells.extend([
            nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("plotter = abilab.ArpesPlotter.pickle_load('%s')" % tmpfile),
            nbv.new_code_cell("print(plotter)"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


if __name__ == "__main__":
    import sys
    plotter = ArpesPlotter.model_from_ebands(sys.argv[1]) #, aw, aw_meshes, tmesh)
    print(plotter.to_string(verbose=2))
    #plotter.plot_ekmap_itemp(itemp=0, estep=0.05, with_colorbar=True)
    plotter.plot_ekmap_temps(estep=0.05, with_colorbar=True)
    #plotter.plot_3dlines(itemp=0, estep=0.05, band_inds=[1, 2, 3])
    #plotter.plot_ak()
    #plotter.plot_ak_vs_temp()
    #plotter.plot_surface(istep=0, estep=0.05)



# coding: utf-8
"""Interface to the wout output file produced by Wannier90."""
from __future__ import annotations

import numpy as np
import pandas as pd

from collections import OrderedDict
from monty.string import marquee
from abipy.core.mixins import BaseFile, Has_Structure, NotebookWriter
from abipy.core.structure import Structure
from abipy.tools.plotting import add_fig_kwargs, get_axarray_fig_plt
from abipy.tools.typing import Figure


class WoutFile(BaseFile, Has_Structure, NotebookWriter):
    """
    Main output file produced by Wannier90

    Usage example:

    .. code-block:: python

        with abilab.abiopen("foo.wout") as wout:
            print(wout)
            wout.plot()

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: WoutFile
    """
    def __init__(self, filepath):
        super().__init__(filepath)
        self.warnings = []
        self.use_disentangle = False
        self.conv_df, self.dis_df = None, None

        with open(self.filepath, "rt") as fh:
            self.lines = fh.readlines()

        self._parse_dims()
        try:
            self._parse_iterations()
        except Exception as exc:
            print("Exception in _parse_iterations:\n", exc)

    def close(self) -> None:
        """Close file. Required by abc protocol."""

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app("Wannier90 version: %s" % self.version)
        app("Number of Wannier functions: %d" % self.nwan)
        app("K-grid: %s" % self.grid_size)
        if self.use_disentangle:
            app("Using DISENTANGLE algorithm")
            #for k, v in self.params_section["DISENTANGLE"].items():
            #    app("%s: %s" % (k, v))
        app("")

        if self.dis_df is not None:
            # Print first and last n cycles.
            app(marquee("DISENTANGLE", mark="="))
            n = 5 if not verbose else 20
            if len(self.dis_df) > 2 * n:
                app(pd.concat([self.dis_df.head(n), self.dis_df.tail(n)]).to_string(index=False))
            else:
                app(self.dis_df.to_string(index=False))
            app("")

        if self.conv_df is not None:
            # Print first and last n cycles.
            app(marquee("WANNIERISE", mark="="))
            n = 5 if not verbose else 20
            if len(self.conv_df) > 2 * n:
                app(pd.concat([self.conv_df.head(n), self.conv_df.tail(n)]).to_string(index=False))
            else:
                app(self.conv_df.to_string(index=False))
            app("")

        if self.warnings:
            app("Found %d warnings in output file:" % len(self.warnings))
            for i, w in enumerate(self.warnings):
                app("[%d] %s" % (i, w))

        #if verbose:

        return "\n".join(lines)

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self._structure

    def _parse_dims(self) -> None:
        """
        Parse basic dimensions and get structure from the header of the file.
        """
        self.version, self._structure, self.grid_size = None, None, None
        # Init dictionary with parameters.
        self.params_section = OrderedDict([(s, OrderedDict()) for s in
            ("MAIN", "WANNIERISE", "PLOTTING", "DISENTANGLE")])
        params_done = False

        for iln, line in enumerate(self.lines):
            # Check for any warnings
            if 'Warning' in line:
                self.warnings.append(line)
                continue

            if "Time to read parameters" in line:
                params_done = True
                continue

            # Get release string.
            if "Release:" in line:
                i = line.find("Release:")
                self.version = line[i:].split()[1]
                continue

            # Parse lattice.
            if "Lattice Vectors" in line and self._structure is None:
                #              Lattice Vectors (Ang)
                #    a_1     0.000000   2.715473   2.715473
                #    a_2     2.715473   0.000000   2.715473
                #    a_3     2.715473   2.715473   0.000000
                lattice = np.array([list(map(float, self.lines[iln+j].split()[1:])) for j in range(1, 4)])
                continue

            # Parse atoms.
            if "|   Site   " in line and self._structure is None:
                # *----------------------------------------------------------------------------*
                # |   Site       Fractional Coordinate          Cartesian Coordinate (Ang)     |
                # +----------------------------------------------------------------------------+
                # | Si   1   0.00000   0.00000   0.00000   |    0.00000   0.00000   0.00000    |
                # | Si   2   0.25000   0.25000   0.25000   |    1.35774   1.35774   1.35774    |
                # *----------------------------------------------------------------------------*
                frac_coords, species = [], []
                i = iln + 2
                while True:
                    l = self.lines[i].strip()
                    if l.startswith("*"): break
                    i += 1
                    tokens = l.replace("|", " ").split()
                    species.append(tokens[0])
                    frac_coords.append(np.array(list(map(float, tokens[2:5]))))

                self._structure = Structure(lattice, species, frac_coords)
                continue

            # Parse kmesh.
            if "Grid size" in line:
                # Grid size =  2 x  2 x  2      Total points =    8
                tokens = line.split("=")[1].split("Total")[0].split("x")
                self.grid_size = np.array(list(map(int, tokens)))
                continue

            if not params_done and any(sname in line for sname in self.params_section):
                #*---------------------------------- MAIN ------------------------------------*
                #|  Number of Wannier Functions               :                 4             |
                #|  Wavefunction spin channel                 :                up             |
                #*----------------------------------------------------------------------------*
                # Use params_done to avoid parsing the second section with WANNIERISE
                key = line.replace("*", "").replace("-", "").strip()
                i = iln + 1
                l = self.lines[i].strip()
                while not l.startswith("*-"):
                    tokens = [s.strip() for s in l.replace("|", "").split(":")]
                    self.params_section[key][tokens[0]] = tokens[1]
                    i += 1
                    l = self.lines[i].strip()
                continue

        # Extract important metadata from sections and convert from string.
        self.nwan = int(self.params_section["MAIN"]["Number of Wannier Functions"])
        if self.params_section["DISENTANGLE"].get("Using band disentanglement", "F") == "T":
            self.use_disentangle = True

    def _parse_iterations(self) -> int:
        """
        Parse iteration steps if not already done and store results in self.

        Return: 0 if success.
        """
        # Don't parse it again if already done.
        if self.conv_df is not None: return 0

        if self.use_disentangle:
            # Parse Disentanglement cycles
            # +---------------------------------------------------------------------+<-- DIS
            # |  Iter     Omega_I(i-1)      Omega_I(i)      Delta (frac.)    Time   |<-- DIS
            # +---------------------------------------------------------------------+<-- DIS
            #       1       3.91743302       3.66269149       6.955E-02      0.28    <-- DIS
            #       2       3.66269149       3.66269149       2.021E-14      0.29    <-- DIS
            # <<<      Delta < 1.000E-10  over  3 iterations     >>>
            # <<< Disentanglement convergence criteria satisfied >>>
            in_diis = 0
            data = OrderedDict([(s, []) for s in
                ("iter", "omegaI_im1", "omegaI_i", "delta_frac", "time")])

            for line in self.lines:
                line = line.strip()
                if not line.endswith("<-- DIS"):
                    if in_diis: break
                    continue
                in_diis += 1
                if in_diis >= 4:
                    toks = line.split()
                    for i, key in enumerate(data.keys()):
                        data[key].append(int(toks[i]) if i == 0 else float(toks[i]))

            self.dis_df = pd.DataFrame.from_dict(data)

        # Parse Wannierization cycles.
        for start, line in enumerate(self.lines):
            if "Initial State" in line: break
        else:
            return 1

        self.wf_centers = [[] for _ in range(self.nwan)]
        self.wf_spreads = [[] for _ in range(self.nwan)]
        data = OrderedDict([(s, []) for s in
            ("iter",  "delta_spread", "rms_gradient", "spread", "time",
             "O_D", "O_OD", "O_TOT",
            )])

        lines = self.lines[start:][:]
        while lines:
            line = lines.pop(0).strip()
            if not line.startswith("Initial State") and not line.startswith("Cycle:"): continue
            step = int(line.split()[-1]) if line.startswith("Cycle:") else 0
            for iw in range(self.nwan + 1):
                # WF centre and spread    1  (  0.042127,  0.071712, -0.424794 )    10.42287858
                # Sum of centres and spreads (  0.933074, -0.071343, -0.800933 )    42.11245002
                tokens = lines.pop(0).replace("(", " ").replace(")", " ").replace(",", "").split()
                spread = float(tokens[-1])
                center = list(map(float, tokens[-4:-1]))
                if iw != self.nwan:
                    self.wf_centers[iw].append(center)
                    self.wf_spreads[iw].append(spread)

            line = lines.pop(0).strip()
            if line:
                raise ValueError("Expecting empty string. Got:\n`%s`" % line)

            #      0     0.421E+02     0.0000000000       42.1124500153       0.57  <-- CONV
            #        O_D=     32.4608805 O_OD=      5.9016636 O_TOT=     42.1124500 <-- SPRD
            # ------------------------------------------------------------------------------
            # Cycle:      1
            #  WF centre and spread    1  ( -0.141379, -0.258009, -0.488755 )     8.73529609
            #  Sum of centres and spreads (  0.794432, -0.178304, -0.823458 )    32.08361376
            #      1    -0.100E+02    10.2774612971       32.0836137609       0.57  <-- CONV
            #        O_D=     22.6617600 O_OD=      5.6719478 O_TOT=     32.0836138 <-- SPRD
            # Delta: O_D= -0.9799120E+01 O_OD= -0.2297158E+00 O_TOT= -0.1002884E+02 <-- DLTA
            # ------------------------------------------------------------------------------
            # Parse CONV and add values to data
            conv_toks = lines.pop(0).split()
            for i, k in enumerate(("iter", "delta_spread", "rms_gradient", "spread", "time")):
                data[k].append(int(conv_toks[i]) if i == 0 else float(conv_toks[i]))

            sprd_toks = lines.pop(0).split()
            data["O_D"].append(float(sprd_toks[1]))
            data["O_OD"].append(float(sprd_toks[3]))
            data["O_TOT"].append(float(sprd_toks[5]))
            if step > 0:
                dlta_tokens = lines.pop(0).split()

        self.conv_df = pd.DataFrame.from_dict(data)

        # Convert to numpy array (nwan, nstep, 3) and (nwan, nstep)
        self.wf_centers = np.array(self.wf_centers)
        self.wf_spreads = np.array(self.wf_spreads)

        return 0

    @add_fig_kwargs
    def plot(self, fontsize=8, **kwargs) -> Figure:
        """
        Plot the convergence of the Wannierise cycle.

        Args:
            fontsize: legend and label fontsize.

        Returns: |matplotlib-Figure|
        """
        if self._parse_iterations() != 0:
            print("Wout files does not contain Wannierization cycles. Returning None")
            return None

        items = ["delta_spread", "rms_gradient", "spread"]
        if self.use_disentangle:
            items += ["omegaI_i"]

        # Build grid of plots.
        num_plots, ncols, nrows = len(items), 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=False, squeeze=False)
        ax_list = ax_list.ravel()

        # Don't show the last ax if num_plots is odd.
        if num_plots % ncols != 0: ax_list[-1].axis("off")

        marker = "."
        for ax, item in zip(ax_list, items):
            ax.grid(True)
            ax.set_xlabel("Iteration Step")
            ax.set_ylabel(item)
            s = 1
            if item == "omegaI_i":
                # Plot Disentanglement cycles
                ax.plot(self.dis_df.iter[s:], self.dis_df[item][s:], marker=marker)
                from mpl_toolkits.axes_grid1.inset_locator import inset_axes
                ax2 = inset_axes(ax, width="60%", height="40%", loc="upper right")
                ax2.grid(True)
                ax2.set_title("delta_frac", fontsize=fontsize)
                ax2.plot(self.dis_df.iter[s:], self.dis_df["delta_frac"][s:], marker=marker)

            else:
                ax.plot(self.conv_df.iter[s:], self.conv_df[item][s:], marker=marker)

        return fig

    @add_fig_kwargs
    def plot_centers_spread(self, fontsize=8, **kwargs) -> Figure:
        """
        Plot the convergence of the Wannier centers and spread
        as a function of the iteration number

        Args:
            fontsize: legend and label fontsize.

        Returns: |matplotlib-Figure|
        """
        if self._parse_iterations() != 0:
            print("Wout files does not contain Wannierization cycles. Returning None")
            return None

        # Build grid of plots.
        # nwan subplot with evolution of the WF center + last subplot with all spreads
        num_plots, ncols, nrows = self.nwan + 1, 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=False, squeeze=False)
        ax_list = ax_list.ravel()

        # Don't show the last ax if num_plots is odd.
        if num_plots % ncols != 0: ax_list[-1].axis("off")

        marker = "."
        for iax in range(num_plots):
            ax = ax_list[iax]
            ax.grid(True)
            ax.set_xlabel("Iteration Step")
            s = 1
            if iax < self.nwan:
                ax.set_ylabel("Center of WF #%s" % (iax + 1))
                for idir in range(3):
                    ax.plot(self.conv_df.iter[s:], self.wf_centers[iax, s:, idir], marker=marker,
                            label={0: "x", 1: "y", 2: "z"}[idir] if iax == 0 else None)
            else:
                ax.set_ylabel("WF Spread")
                for iw in range(self.nwan):
                    ax.plot(self.conv_df.iter[s:], self.wf_spreads[iw, s:], marker=marker,
                            label="WF#%d" % (iw + 1))

            if iax in (0, self.nwan):
                ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.plot(show=False)
        yield self.plot_centers_spread(show=False)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("wout = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(wout)"),
            nbv.new_code_cell("wout.structure.plot();"),
            nbv.new_code_cell("wout.plot();"),
            nbv.new_code_cell("wout.plot_centers_spread();"),
        ])

        return self._write_nb_nbpath(nb, nbpath)



#!/usr/bin/env python
r"""
X-ray diffraction pattern
=========================

This example shows how to plot the X-ray diffraction pattern with pymatgen
"""
from abipy import abilab
import abipy.data as abidata

# Extract the structure (supports multiple formats e.g netcdf, Abinit input, cif, POSCAR)
# Also available via the `abistruct.py xrd FILE` command line interface.
structure = abilab.Structure.from_file(abidata.ref_file("si_scf_WFK.nc"))

import sys
if sys.version[0:3] > '2.7':
    # pmg broke py compatibility
    structure.plot_xrd()



# coding: utf-8
"""Objects to analyze the screening files produced by the GW code (optdriver 3)."""
from __future__ import annotations

import numpy as np
import pymatgen.core.units as pmgu

from monty.string import marquee
from monty.inspect import all_subclasses
from monty.termcolor import cprint
from monty.collections import AttrDict
from monty.functools import lazy_property
from monty.bisect import index as bs_index
from abipy.core.func1d import Function1D
from abipy.core.kpoints import KpointList
from abipy.core.structure import Structure
from abipy.core.gsphere import GSphere
from abipy.core.mixins import AbinitNcFile, Has_Header, Has_Structure, NotebookWriter
from abipy.electrons.ebands import ElectronBands
from abipy.iotools import ETSF_Reader
from abipy.tools.plotting import ArrayPlotter, data_from_cplx_mode, add_fig_kwargs, get_ax_fig_plt, set_axlims
from abipy.tools.typing import Figure
from abipy.tools import duck


_COLOR_CMODE = dict(re="red", im="blue", abs="black", angle="green")


def _latex_symbol_cplxmode(symbol: str, cplx_mode: str) -> str:
    """Latex label to be used to plot ``symbol`` in ``cplx_mode``."""
    return {"re": r"$\Re(" + symbol + ")$",
            "im": r"$\Im(" + symbol + ")$",
            "abs": r"$||" + symbol + "||$",
            "angle": r"$Phase(" + symbol + ")$"}[cplx_mode]


# TODO: Should contain ElectronBands.
class ScrFile(AbinitNcFile, Has_Header, Has_Structure, NotebookWriter):
    """
    This object provides an interface to the ``SCR.nc`` file produced by the GW code.

    Usage example:

    .. code-block:: python

        with ScrFile("foo_SCR.nc") as ncfile:
            print(ncfile)
            ncfile.plot_emacro()

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: ScrFile
    """
    @classmethod
    def from_file(cls, filepath: str) -> ScrFile:
        """Initialize the object from a Netcdf file"""
        return cls(filepath)

    def __init__(self, filepath: str):
        super().__init__(filepath)
        self.reader = ScrReader(filepath)

    def close(self) -> None:
        """Close the file."""
        self.reader.close()

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        # TODO: Fix problem with efermi
        #app(self.ebands.to_string(with_structure=False, title="Electronic Bands"))
        app(self.kpoints.to_string(verbose=verbose, title="K-points for screening function"))
        app("")
        app("Number of G-vectors in screening matrices: %d" % self.ng)
        app("Number of frequencies: %d (real: %d, imaginary: %d)" % (self.nw, self.nrew, self.nimw))

        if verbose:
            app(str(self.params))

        if verbose > 1:
            app("")
            app(self.hdr.to_string(verbose=verbose, title="Abinit Header"))

        return "\n".join(lines)

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.reader.structure

    @property
    def kpoints(self):
        """List of k-points in the dielectric function."""
        return self.reader.kpoints

    @lazy_property
    def ebands(self) -> ElectronBands:
        """
        |ElectronBands| object with the single-particle energies used to compute the screening.
        """
        ebands = ElectronBands.from_file(self.filepath)
        # FIXME
        cprint("Setting Fermi energy to zero since `fermie_energy` is not initialized in Abinit v8.2", "yellow")
        ebands.fermie = 0
        return ebands

    @property
    def ng(self) -> int:
        """Number of G-vectors in screening matrices."""
        return self.reader.ng

    @property
    def wpoints(self) -> np.ndarray:
        """
        Array of complex numbers with the frequencies of the dielectric function in Hartree.
        """
        return self.reader.wpoints

    @property
    def nw(self) -> int:
        """Total number of frequencies."""
        return self.reader.nw

    @property
    def nrew(self) -> int:
        """Number of real frequencies."""
        return self.reader.nrew

    @property
    def nimw(self) -> int:
        """Number of imaginary frequencies."""
        return self.reader.nimw

    @property
    def netcdf_name(self) -> str:
        """The netcdf_ name associated to the data on disk."""
        return self.reader.netcdf_name

    @lazy_property
    def params(self) -> dict:
        """
        dict with the most important parameters used to compute the screening
        keys can be accessed with the dot notation i.e. ``params.zcut``.
        """
        #od = self.get_ebands_params()
        return self.reader.read_params()

    @add_fig_kwargs
    def plot_emacro(self, cplx_mode="re-im", ax=None, xlims=None, fontsize=12, **kwargs) -> Figure:
        r"""
        Plot the macroscopic dielectric function with local-field effects.

        Args:
            cplx_mode: string defining the data to print.
                Possible choices are (case-insensitive):
                "re" for the real part. "im" for the imaginary part.
                "abs" for the absolute value.
                "angle" will display the phase of the complex number in radians.
                Options can be concatenated with ``-`` e.g. ``re-im``.
            xlims: Set the data limits for the x-axis in eV. Accept tuple e.g. (left, right)
                or scalar e.g. left. If left (right) is None, default values are used.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: Legend and title fontsize.

        Returns:
            |matplotlib-Figure|
        """
        emlf = self.reader.read_emacro_lf()
        xx, yy = emlf.mesh * pmgu.Ha_to_eV, emlf.values

        ax, fig, plt = get_ax_fig_plt(ax=ax)
        for c in cplx_mode.lower().split("-"):
            ax.plot(xx, data_from_cplx_mode(c, yy),
                    color=_COLOR_CMODE[c], linewidth=kwargs.get("linewidth", 2),
                    linestyle=kwargs.get("linestyle", "solid"),
                    label=_latex_symbol_cplxmode(r"\varepsilon_{M}", c))

        set_axlims(ax, xlims, "x")
        ax.grid(True)
        ax.set_xlabel(r"$\omega$ (eV)")
        ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_eelf(self, ax=None, xlims=None, fontsize=12, **kwargs) -> Figure:
        r"""
        Plot electron energy loss function.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            xlims: Set the data limits for the x-axis in eV. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used
            fontsize: Legend and label fontsize:

        Returns: |matplotlib-Figure|
        """
        eelf = self.reader.read_eelf()
        xx, yy = eelf.mesh * pmgu.Ha_to_eV, eelf.values

        ax, fig, plt = get_ax_fig_plt(ax=ax)
        ax.plot(xx, yy, linewidth=kwargs.get("linewidth", 2),
                linestyle=kwargs.get("linestyle", "solid"), label="EELF")

        set_axlims(ax, xlims, "x")
        ax.grid(True)
        ax.set_xlabel(r"$\omega$ (eV)")
        ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        edos = self.ebands.get_edos()
        yield self.ebands.plot_with_edos(edos, show=False)
        # Plot spectra if there are enough frequencies.
        if self.nrew > 2:
            yield self.plot_emacro(show=False)
            yield self.plot_eelf(show=False)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to nbpath. If ``nbpath`` is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("ncfile = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(ncfile)"),
            nbv.new_code_cell("print(ncfile.params)"),
            #nbv.new_code_cell("ncfile.ebands.plot();"),
            nbv.new_code_cell("edos = ncfile.ebands.get_edos()\nncfile.ebands.plot_with_edos(edos);"),
        ])

        if self.nrew > 2:
            # Plot optical properties and EELF
            nb.cells.extend([
                nbv.new_code_cell("ncfile.plot_emacro();"),
                nbv.new_code_cell("ncfile.plot_eelf();"),
            ])

        return self._write_nb_nbpath(nb, nbpath)


class SusFile(ScrFile):
    pass


class ScrReader(ETSF_Reader):
    """
    This object reads the results stored in the SCR (Screening) file produced by ABINIT.
    It provides helper functions to access the most important quantities.

    double inverse_dielectric_function(number_of_qpoints_dielectric_function,
    number_of_frequencies_dielectric_function, number_of_spins, number_of_spins,
    number_of_coefficients_dielectric_function, number_of_coefficients_dielectric_function, complex)

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: ScrReader
    """
    def __init__(self, filepath: str):
        super().__init__(filepath)

        # Read and store important quantities.
        self.structure = self.read_structure()
        qfrac_coords = self.read_value("qpoints_dielectric_function")
        self.kpoints = KpointList(self.structure.reciprocal_lattice, qfrac_coords)
        self.wpoints = self.read_value("frequencies_dielectric_function", cmode="c")
        self.ng = self.read_dimvalue("number_of_coefficients_dielectric_function")

        # Find number of real/imaginary frequencies.
        self.nw = len(self.wpoints)
        self.nrew = self.nw
        self.nimw = 0
        for i, w in enumerate(self.wpoints):
            if np.iscomplex(w):
                self.nrew = i
                break

        self.nimw = self.nw - self.nrew
        if self.nimw and not np.all(np.iscomplex(self.wpoints[self.nrew+1:])):
            raise ValueError("wpoints should contained real points packed in the first positions\n"
                             "followed by imaginary points but got: %s" % str(self.wpoints))

        # Define self.netcdf_name from the data available on file.
        nfound = 0
        netcdf_names = ["polarizability", "dielectric_function", "inverse_dielectric_function"]
        for tryname in netcdf_names:
            if tryname in self.rootgrp.variables:
                self.netcdf_name = tryname
                nfound += 1

        if nfound == 0:
            raise RuntimeError("Cannot find `%s` in netcdf file" % str(netcdf_names))
        if nfound > 1:
            raise RuntimeError("Find multiple netcdf arrays (%s) in netcdf file!" % str(netcdf_names))

    def read_params(self) -> AttrDict:
        """
        Read the most important parameters used to compute the screening i.e.
        the parameters that may be subject to convergence studies.

        Returns:
            |AttrDict| a dictionary whose keys can be accessed with the dot notation i.e. ``d.key``.
        """
        # TODO: ecuteps is missing!
        keys = ["ikxc", "inclvkb", "gwcalctyp", "nbands_used", "npwwfn_used",
                "spmeth", "test_type", "tordering", "awtr", "icutcoul", "gwcomp",
                "gwgamma", "mbpt_sciss", "spsmear", "zcut", "gwencomp"]

        def convert(arr):
            """Convert to scalar if size == 1"""
            #return np.asscalar(arr) if arr.size == 1 else arr
            return np.asarray(arr).item() if arr.size == 1 else arr

        return AttrDict({k: convert(self.read_value(k)) for k in keys})

    def read_emacro_lf(self, kpoint=(0, 0, 0)) -> Function1D:
        """
        Read the macroscopic dielectric function *with* local field effects 1 / em1_{0,0)(kpoint, omega).

        Return: |Function1D| object.
        """
        if self.netcdf_name == "inverse_dielectric_function":
            em1 = self.read_wslice(kpoint, ig1=0, ig2=0)
            emacro = 1 / em1[:self.nrew]
        else:
            raise NotImplementedError("emacro_lf with netcdf != InverseDielectricFunction")

        return Function1D(np.real(self.wpoints[:self.nrew]).copy(), emacro)

    def read_emacro_nlf(self, kpoint=(0, 0, 0)) -> Function1D:
        """
        Read the macroscopic dielectric function *without* local field effects e_{0,0)(kpoint, omega).

        Return: |Function1D|

        .. warning::

            This function performs the inversion of e-1 to get e.
            that can be quite expensive and memory demanding for large matrices!
        """
        if self.netcdf_name == "inverse_dielectric_function":
            em1 = self.read_wggmat(kpoint)
            e = np.linalg.inv(em1.wggmat[:self.nrew, :, :])
        else:
            raise NotImplementedError("emacro_nlf with netcdf != InverseDielectricFunction")

        return Function1D(np.real(self.wpoints[:self.nrew]).copy(), e[:, 0, 0])

    def read_eelf(self, kpoint=(0, 0, 0)) -> Function1D:
        """
        Read electron energy loss function

            - Im(1/ emacro)

        Return: |Function1D| object.
        """
        # eelf = -Im(1 / eM)
        emacro_lf = self.read_emacro_lf(kpoint=kpoint)
        #emacro_lf = self.read_emacro_nlf(kpoint=kpoint)
        values = (-1 / emacro_lf.values).imag

        return Function1D(emacro_lf.mesh.copy(), values)

    def read_wggmat(self, kpoint, spin1=0, spin2=0, cls=None):
        """
        Read data at the given k-point and return an instance of ``cls`` where
        ``cls`` is a subclass of :class:`_AwggMatrix`
        """
        cls = _AwggMatrix.class_from_netcdf_name(self.netcdf_name) if cls is None else cls

        var = self.rootgrp.variables["reduced_coordinates_plane_waves_dielectric_function"]
        # Use ik = 0 because the basis set is not k-dependent.
        ik0 = 0
        gvecs = var[ik0, :]
        #print("gvecs", gvecs)

        kpoint, ik = self.find_kpoint_fileindex(kpoint)

        # FIXME ecuteps is missing
        # TODO: Gpshere.find is very slow if we don't take advantage of shells
        ecuteps = 2
        gsphere = GSphere(ecuteps, self.structure.reciprocal_lattice, kpoint, gvecs)

        # Exchange spin due to F --> C
        values = self.rootgrp.variables[self.netcdf_name][ik, :, spin2, spin1, :, :, :]
        wggmat = values[:, :, :, 0] + 1j * values[:, :, :, 1]

        return cls(self.wpoints, gsphere, wggmat, inord="F")

    def find_kpoint_fileindex(self, kpoint):
        """
        Returns the k-point and the index of the k-point in the netcdf file.
        Accepts |Kpoint| instance or integer.
        """
        if duck.is_intlike(kpoint):
            ik = int(kpoint)
        else:
            ik = self.kpoints.index(kpoint)

        return self.kpoints[ik], ik

    def read_wslice(self, kpoint, ig1=0, ig2=0, spin1=0, spin2=0):
        """Read slice along the frequency dimension."""
        kpoint, ik = self.find_kpoint_fileindex(kpoint)
        var = self.rootgrp.variables[self.netcdf_name]
        values = var[ik, :, spin2, spin1, ig2, ig1, :]  # Exchange G indices F --> C

        return values[:, 0] + 1j * values[:, 1]


class _AwggMatrix:
    r"""
    Base class for two-point functions expressed in reciprocal space
    i.e. a complex matrix :math:`A_{G,G'}(\omega)` where G, G' are reciprocal
    lattice vectors defined inside the G-sphere.

    This class is not supposed to be instantiated directly.

    .. attributes:

        wpoints:
        gpshere:
        nrew:
        nwim:
    """
    netcdf_name = "_AwggMatrix"

    latex_name = "Unknown"

    def __init__(self, wpoints, gsphere, wggmat, inord="C"):
        """"
        Args:
            gsphere: |GSphere| with G-vectors and k-point object.
            wpoints: Complex frequency points in Hartree.
            wggmat: [nw, ng, ng] complex array.
            inord: storage order of ``wggmat``. If inord == "F", ``wggmat`` is in
                in Fortran column-major order. Default: "C" i.e. C row-major order.
        """
        self.wpoints = np.array(wpoints, dtype=complex)
        self.gsphere = gsphere
        self.wggmat = np.reshape(wggmat, (self.nw, self.ng, self.ng))

        if inord.lower() == "f":
            # Fortran to C.
            for iw, _ in enumerate(wpoints):
                self.wggmat[iw] = self.wggmat[iw].T.copy()

        for i in (1, 2):
            assert len(gsphere) == wggmat.shape[-i]
        assert len(self.wpoints) == len(self.wggmat)

        # Find number of real/imaginary frequencies.
        self.nrew = self.nw
        self.nimw = 0
        for i, w in enumerate(self.wpoints):
            if np.iscomplex(w):
                self.nrew = i
                break

        self.nimw = self.nw - self.nrew
        if self.nimw and not np.all(np.iscomplex(self.wpoints[self.nrew+1:])):
            raise ValueError("wpoints should contained real points packed in the first positions\n"
                "followed by imaginary points but got: %s" % str(self.wpoints))

    @classmethod
    def class_from_netcdf_name(cls, netcdf_name: str):
        """Return the subclass associated to the given netcdf name."""
        nfound = 0
        for subclass in all_subclasses(cls):
            if subclass.netcdf_name == netcdf_name:
                nfound += 1
                out_cls = subclass

        if nfound == 0:
            raise ValueError("Cannot find subclass associated to `%s`" % str(netcdf_name))
        if nfound > 1:
            raise ValueError("Find multiple subclasses associated to `%s`" % str(netcdf_name))

        return out_cls

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation."""
        lines = []
        app = lines.append

        app(marquee(self.netcdf_name, mark="="))
        app("  K-point: %s" % self.kpoint)
        app("  Number of G-vectors: %d" % self.ng)
        app("  Total number of frequencies: %d (real: %s, imaginary: %s)" % (self.nw, self.nrew, self.nimw))
        if self.nrew:
            app("  Real frequencies up to %.2f (eV)" % self.real_wpoints[-1].real)
        if self.nimw:
            app("  Imaginary frequencies up to %.2f (eV)" % self.imag_wpoints[-1].imag)

        return "\n".join(lines)

    @property
    def kpoint(self):
        """|Kpoint| object."""
        return self.gsphere.kpoint

    @property
    def ng(self) -> int:
        """Number of G-vectors."""
        return len(self.gsphere)

    @property
    def nw(self) -> int:
        """Total number of frequencies."""
        return len(self.wpoints)

    @property
    def real_wpoints(self):
        """
        Real frequencies in Hartree. Empty list if not available.
        """
        if self.nrew > 0:
            return np.real(self.wpoints[:self.nrew])
        return []

    @property
    def imag_wpoints(self):
        """
        Imaginary frequencies in Hartree. Empty list if not available.
        """
        if self.nimw > 0:
            return self.wpoints[self.nrew:]
        return []

    @property
    def wggmat_realw(self):
        """The slice of wggmat along the real axis."""
        return self.wggmat[:self.nrew, :, :]

    @property
    def wggmat_imagw(self):
        """The slice of wggmat along the imaginary axis."""
        return self.wggmat[self.nrew:, :, :]

    def windex(self, w, atol=0.001):
        """
        Find the index of the **closest** frequency in ``wpoints``.
        """
        if np.iscomplex(w):
            iw = bs_index(self.imag_wpoints.imag, w.imag, atol=atol)
            iw += self.nrew
        else:
            iw = bs_index(self.real_wpoints.real, w, atol=atol)

        return iw

    def gindex(self, gvec):
        """
        Find the index of gvec. If ``gvec`` is an integer, gvec is returned.
        Raises:
            `ValueError` if gvec is not found.
        """
        if duck.is_intlike(gvec): return int(gvec)
        return self.gsphere.index(gvec)

    def latex_label(self, cplx_mode):
        """Return a latex string that can be used in matplotlib plots."""
        return _latex_symbol_cplxmode(self.latex_name, cplx_mode)

    @add_fig_kwargs
    def plot_freq(self, gvec1, gvec2=None, waxis="real", cplx_mode="re-im", ax=None, fontsize=8, **kwargs) -> Figure:
        r"""
        Plot the frequency dependence of :math:`W_{G1, G2}(\omega)`

        Args:
            gvec1, gvec2:
            waxis: ``real`` to plot along the real axis, ``imag`` for the imaginary axis.
            cplx_mode: string defining the data to print.
                Possible choices are (case-insensitive): ``re`` for the real part
                ``im`` for the imaginary part, ``abs`` for the absolute value.
                ``angle`` will display the phase of the complex number in radians.
                Options can be concatenated with ``-`` e.g. ``re-im``.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: legend and label fontsize.

        Returns: |matplotlib-Figure|
        """
        # Select data to plot
        ig1 = self.gindex(gvec1)
        ig2 = ig1 if gvec2 is None else self.gindex(gvec2)

        ax, fig, plt = get_ax_fig_plt(ax=ax)
        if waxis == "real":
            if self.nrew == 0: return fig
            xx = self.real_wpoints.real * pmgu.Ha_to_eV
            yy = self.wggmat_realw[:, ig1, ig2]

        elif waxis == "imag":
            if self.nimw == 0: return fig
            xx = self.imag_wpoints.imag * pmgu.Ha_to_eV
            yy = self.wggmat_imagw[:, ig1, ig2]

        else:
            raise ValueError("Wrong value for waxis: %s" % str(waxis))

        linewidth = kwargs.pop("linewidth", 2)
        linestyle = kwargs.pop("linestyle", "solid")

        lines = []
        for c in cplx_mode.lower().split("-"):
            l, = ax.plot(xx, data_from_cplx_mode(c, yy),
                         color=_COLOR_CMODE[c], linewidth=linewidth, linestyle=linestyle,
                         label=self.latex_label(c))
            lines.append(l)

        ax.grid(True)
        ax.set_xlabel(r"$\omega$ (eV)")
        ax.set_title("%s, kpoint: %s" % (self.netcdf_name, self.kpoint), fontsize=fontsize)
        ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    @add_fig_kwargs
    def plot_gg(self, cplx_mode="abs", wpos=None, **kwargs) -> Figure:
        r"""
        Use matplotlib imshow to plot :math:`W_{GG'}` matrix

        Args:
            cplx_mode: string defining the data to print.
                Possible choices are (case-insensitive): ``re`` for the real part
                ``im`` for the imaginary part, ``abs`` for the absolute value.
                ``angle`` will display the phase of the complex number in radians.
            wpos: List of frequency indices to plot. If None, the first frequency is used (usually w=0).
                If wpos == "all" all frequencies are shown (use it carefully)
                Other possible values: "real" if only real frequencies are wanted.
                "imag" for imaginary frequencies only.

        Returns: |matplotlib-Figure|
        """
        # Get wpos indices.
        choice_wpos = {None: [0], "all": range(self.nw),
                       "real": range(self.nrew), "imag": range(self.nrew, self.nw)}

        if any(wpos == k for k in choice_wpos):
            wpos = choice_wpos[wpos]
        else:
            if duck.is_intlike(wpos): wpos = [int(wpos)]
            wpos = np.array(wpos)

        # Build plotter.
        plotter = ArrayPlotter()
        for iw in wpos:
            label = r"%s $\omega=%s$" % (self.latex_label(cplx_mode), self.wpoints[iw])
            data = data_from_cplx_mode(cplx_mode, self.wggmat[iw])
            plotter.add_array(label, data)

        return plotter.plot(show=False, **kwargs)


class Polarizability(_AwggMatrix):
    """
    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: Polarizability
    """
    netcdf_name = "polarizability"
    latex_name = r"\tilde chi"


class DielectricFunction(_AwggMatrix):
    """
    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: DielectricFunction
    """
    netcdf_name = "dielectric_function"
    latex_name = r"\epsilon"


class InverseDielectricFunction(_AwggMatrix):
    """
    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: InverseDielectricFunction
    """
    netcdf_name = "inverse_dielectric_function"
    latex_name = r"\epsilon^{-1}"

    #def _add_ppmodel(self, ppm):
    #    """
    #    Add a :class:`PPModel` object to the internal list. Return ppm.
    #    """
    #    if not hasattr(self, "ppmodels"): self.ppmodels = []
    #    self.ppmodels.append(ppm)
    #    return ppm

    #def build_godby_needs_ppmodel(self, wplasma):
    #    ppm = GodbyNeeds.from_em1(self, wplasma)
    #    return self._add_ppmodel(ppm)

    #@add_fig_kwargs
    #def plot_with_ppmodels(self, gvec1, gvec2=None, waxis="real", cplx_mode="re",
    #                       zcut=0.1/pmgu.Ha_to_eV, **kwargs):
    #    """
    #    Args:
    #        gvec1, gvec2:
    #        waxis: "real" to plot along the real axis, "imag" for the imaginary axis.
    #        cplx_mode: string defining the data to print.
    #           Possible choices are (case-insensitive): `re` for the real part
    #           "im" for the imaginary part, "abs" for the absolute value.
    #           "angle" will display the phase of the complex number in radians.
    #           Options can be concatenated with "-" e.g. "re-im"
    #        zcut: Small shift along the imaginary axis to avoid poles. 0.1 eV is the Abinit default.
    #        ax: matplotlib :class:`Axes` or None if a new figure should be created.

    #    Returns:
    #        matplotlib figure.
    #    """
    #    # Select the (G,G') indices to plot.
    #    ig1 = self.gindex(gvec1)
    #    ig2 = ig1 if gvec2 is None else self.gindex(gvec2)

    #    ax, fig, plt = get_ax_fig_plt(ax=None)

    #    self.plot_freq(gvec1, gvec2=gvec2, waxis=waxis, cplx_mode=cplx_mode, ax=ax, show=False)

    #    # Compute em1 from the ppmodel on the same grid used for self.
    #    omegas = {"real": self.real_wpoints, "imag": self.imag_wpoints}[waxis]

    #    # Get y-limits of the ab-initio em1 to zoom-in the interesting region
    #    ymin_em1, ymax_em1 = ax.get_ylim()

    #    for ppm in self.ppmodels:
    #        em1_ppm = ppm.eval_em1(omegas, zcut)
    #        em1_ppm.plot_freq(ig1, gvec2=ig2, waxis=waxis, cplx_mode=cplx_mode,
    #                          ax=ax, linestyle="--", show=False)

    #    ax.set_ylim(ymin_em1, ymax_em1)

    #    return fig


#class PPModel(metaclas=abc.ABCMeta):
#    """
#    Abstract base class for Plasmonpole models.
#    """
#
#    #@abc.abstractmethod
#    #def from_em1(cls, em1):
#    #    """Compute the plasmon-pole parameters from the inverse dielectric function."""
#
#    @abc.abstractmethod
#    def eval_em1(self, omegas, zcut):
#        """Compute the plasmon-pole model at frequency omega (Ha units)."""
#
#
#class GodbyNeeds(PPModel):
#
#    def __init__(self, gsphere, omegatw, bigomegatwsq):
#        r"""
#        bigomegatwsq(:)
#        Plasmon pole parameters $\tilde\Omega^2_{G Gp}(q)$.
#
#        omegatw(:)
#        omegatw(nqibz)%value(npwc,dm2_otq)
#        Plasmon pole parameters $\tilde\omega_{G Gp}(q)$.
#        """
#        self.gsphere = gsphere
#        self.kpoint = gsphere.kpoint
#        self.omegatw = omegatw
#        self.bigomegatwsq = bigomegatwsq
#
#        self.ng = len(gsphere)
#        assert len(omegatw) == len(bigomegatwsq)
#        assert len(omegatw) == len(gsphere)
#
#    @classmethod
#    def from_em1(cls, em1, wplasma):
#        # Find omega=0 and the second imaginary frequency to fit the ppm parameters.
#        iw0 = -1; iw1 = -1
#        for i, w in enumerate(em1.wpoints):
#            if np.abs(w) <= 1e-6: iw0 = i
#            if np.abs(w - 1j*wplasma) <= 1e-6: iw1 = i
#        if iw0 == -1:
#            raise ValueError("Cannot find omega=0 in em1")
#        if iw1 == -1:
#            raise ValueError("Cannot find second imaginary frequency at %s in em1!" % wplasma)
#
#        w0gg = em1.wggmat[iw0, :, :]
#        w1gg = em1.wggmat[iw1, :, :]
#
#        aa = w0gg - np.eye(em1.ng)
#        diff = w0gg - w1gg
#        ratio = aa / diff
#        omegatwsq = (ratio - 1.0) * (wplasma ** 2)
#
#        # If omega-twiddle-squared is negative,set omega-twiddle-squared to 1.0 (a reasonable way of treating
#        # such terms, in which epsilon**-1 was originally increasing along this part of the imaginary axis)
#        # (note: originally these terms were ignored in Sigma; this was changed on 6 March 1990.)
#        #if (REAL(omegatwsq) <= 0.0) omegatwsq=one
#        omegatwsq[np.where(omegatwsq.real <= 0.0)] = 1.0
#        #
#        # Get omega-twiddle
#        # * Neglect the imag part (if any) in omega-twiddle-squared
#        #omegatw(ig,igp)=SQRT(REAL(omegatwsq))
#        #omegatw = np.sqrt(omegatwsq)
#        omegatw = np.sqrt(omegatwsq.real)
#        #omegatw = omegatw + 7j * pmgu.eV_to_Ha
#
#        bigomegatwsq = -aa * omegatw**2
#
#        return cls(em1.gsphere, omegatw, bigomegatwsq)
#
#    def eval_em1(self, omegas, zcut):
#        omegas = np.array(omegas)
#
#        wggmat = np.empty((len(omegas), self.ng, self.ng), dtype=complex)
#        for i, w in enumerate(omegas):
#            # Add shift but only along the real axis.
#            delta = 0.0 if w.imag != 0 else 1j * zcut
#            den = w**2 - np.real((self.omegatw - delta)**2)
#            em1gg = np.eye(self.ng) + self.bigomegatwsq / den
#            #arg = (w - self.omegatw.real) / (0.01 * pmgu.eV_to_Ha)
#            #em1gg = em1gg * (1 - np.exp(-arg**2))
#            wggmat[i] = em1gg
#
#        return InverseDielectricFunction(self.kpoint, omegas, self.gsphere, wggmat)
#
#    @add_fig_kwargs
#    def plot_ggparams(self, **kwargs):
#        plotter = ArrayPlotter(*[
#            (r"$\tilde\omega_{G G'}$", self.omegatw),
#            (r"$\tilde\Omega^2_{G, G'}$", self.bigomegatwsq)])
#
#        return plotter.plot(show=False, **kwargs)
